@@ -1,6 +1,6 @@
 /*
     libparted - a library for manipulating disk partitions
-    Copyright (C) 1999, 2000, 2001 Free Software Foundation, Inc.
+    Copyright (C) 1998, 1999, 2000, 2001, 2002 Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -22,9 +22,8 @@
 #ifndef PED_ENDIAN_H_INCLUDED
 #define PED_ENDIAN_H_INCLUDED
 
-#include <sys/types.h>
 #include "config.h"
-
+#include <stdint.h>
 
 /* returns the n'th least significant byte */
 #define _GET_BYTE(x, n)		( ((x) >> (8 * (n))) & 0xff )
@@ -46,9 +45,9 @@
 				+ (_GET_BYTE(x, 6) << 8)	\
 				+ (_GET_BYTE(x, 7) << 0) )
 
-#define PED_SWAP16(x)		((__u16) _PED_SWAP16( (__u16) x ))
-#define PED_SWAP32(x)		((__u32) _PED_SWAP32( (__u32) x ))
-#define PED_SWAP64(x)		((__u64) _PED_SWAP64( (__u64) x ))
+#define PED_SWAP16(x)		((uint16_t) _PED_SWAP16( (uint16_t) x ))
+#define PED_SWAP32(x)		((uint32_t) _PED_SWAP32( (uint32_t) x ))
+#define PED_SWAP64(x)		((uint64_t) _PED_SWAP64( (uint64_t) x ))
 
 #ifdef WORDS_BIGENDIAN
 
@@ -63,7 +62,7 @@
 #define PED_BE16_TO_CPU(x)	(x)
 #define PED_LE32_TO_CPU(x)	PED_SWAP32(x)
 #define PED_BE32_TO_CPU(x)	(x)
-#define PED_LE64_TO_CPU(x)	PED_SWAP32(x)
+#define PED_LE64_TO_CPU(x)	PED_SWAP64(x)
 #define PED_BE64_TO_CPU(x)	(x)
 
 #else /* !WORDS_BIGENDIAN */
