@@ -1,6 +1,6 @@
 /*
     libparted - a library for manipulating disk partitions
-    Copyright (C) 1998-2000 Free Software Foundation, Inc.
+    Copyright (C) 1999, 2000, 2001 Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -24,12 +24,9 @@
 extern "C" {
 #endif
 
-#define DEBUG
-// #define VERBOSE
-// #define READONLY
+typedef struct _PedArchitecture PedArchitecture;
 
 #include <parted/constraint.h>
-#include <parted/debug.h>
 #include <parted/device.h>
 #include <parted/disk.h>
 #include <parted/exception.h>
@@ -38,8 +35,18 @@ extern "C" {
 
 #include <stdlib.h>
 
-extern int ped_init ();
-extern void ped_done ();
+struct _PedArchitecture {
+	PedDiskArchOps*		disk_ops;
+	PedDeviceArchOps*	dev_ops;
+};
+
+extern const PedArchitecture*	ped_architecture;
+
+/* the architecture can't be changed if there are any PedDevice's.
+ * i.e. you should only be doing this if it's the FIRST thing you do...
+ */
+extern int ped_set_architecture (const PedArchitecture* arch);
+
 extern const char* ped_get_version ();
 
 extern void* ped_malloc (size_t size);
