@@ -1058,12 +1058,6 @@ _disk_check_sanity (PedDisk* disk)
 	for (walk = disk->part_list; walk; walk = walk->next) {
 		PED_ASSERT (!(walk->type & PED_PARTITION_LOGICAL), return 0);
 		PED_ASSERT (!walk->prev || walk->prev->next == walk, return 0);
-
-		if (!strcasecmp (disk->type->name, "msdos")
-			&& disk->dev->geom_known
-			&& walk->type & PED_PARTITION_FREESPACE)
-			PED_ASSERT (walk->geom.length > 5 * disk->dev->sectors,
-				    return 0);
 	}
 
 	if (!ped_disk_extended_partition (disk))
@@ -1668,6 +1662,7 @@ ped_disk_get_max_partition_geometry (PedDisk* disk, PedPartition* part,
 
 	PED_ASSERT(disk != NULL, return NULL);
 	PED_ASSERT(part != NULL, return NULL);
+	PED_ASSERT(ped_partition_is_active (part), return NULL);
 
 	old_geom = part->geom;
 	if (!ped_disk_maximize_partition (disk, part, constraint))
