@@ -1,6 +1,6 @@
 /*
     libparted - a library for manipulating disk partitions
-    Copyright (C) 1999, 2000, 2001, 2002 Free Software Foundation, Inc.
+    Copyright (C) 1999, 2000, 2001 Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -20,13 +20,12 @@
 #include "config.h"
 
 #include <parted/parted.h>
+#include <parted/debug.h>
 #include <parted/endian.h>
-#include <parted/disk_dos.h>
-#include <parted/disk_pc98.h>
 #include <string.h>
 
-#include <libintl.h>
 #if ENABLE_NLS
+#  include <libintl.h>
 #  define _(String) dgettext (PACKAGE, String)
 #else
 #  define _(String) (String)
@@ -52,231 +51,114 @@ static char MBR_BOOT_CODE[] = {
 #define MSDOS_MAGIC		0xAA55
 #define PARTITION_MAGIC_MAGIC	0xf6f6
 
+#define PARTITION_EMPTY		0x00
+#define PARTITION_FAT12		0x01
+#define PARTITION_FAT16_SM	0x04
+#define PARTITION_EXT		0x05
+#define PARTITION_FAT16		0x06
+#define PARTITION_NTFS		0x07
+#define PARTITION_HPFS		0x07
+#define PARTITION_FAT32		0x0b
+#define PARTITION_FAT32_LBA	0x0c
+#define PARTITION_FAT16_LBA	0x0e
+#define PARTITION_EXT_LBA	0x0f
+
+#define PART_FLAG_HIDDEN	0x10	/* Valid for FAT/NTFS only */
+#define PARTITION_FAT12_H	(PARTITION_FAT12	| PART_FLAG_HIDDEN)
+#define PARTITION_FAT16_SM_H	(PARTITION_FAT16_SM	| PART_FLAG_HIDDEN)
+#define PARTITION_EXT_H		(PARTITION_EXT		| PART_FLAG_HIDDEN)
+#define PARTITION_FAT16_H	(PARTITION_FAT16	| PART_FLAG_HIDDEN)
+#define PARTITION_NTFS_H	(PARTITION_NTFS		| PART_FLAG_HIDDEN)
+#define PARTITION_FAT32_H	(PARTITION_FAT32	| PART_FLAG_HIDDEN)
+#define PARTITION_FAT32_LBA_H	(PARTITION_FAT32_LBA	| PART_FLAG_HIDDEN)
+#define PARTITION_FAT16_LBA_H	(PARTITION_FAT16_LBA	| PART_FLAG_HIDDEN)
+
+#define PARTITION_COMPAQ_DIAG	0x12
+#define PARTITION_LINUX_SWAP	0x82
+#define PARTITION_LINUX		0x83
+#define PARTITION_LINUX_EXT	0x85
+#define PARTITION_LINUX_LVM	0x8e
+#define PARTITION_DELL_DIAG	0xde
+#define PARTITION_GPT		0xee
+#define PARTITION_LINUX_RAID	0xfd
+#define PARTITION_LINUX_LVM_OLD 0xfe
+
 typedef struct _DosRawPartition		DosRawPartition;
 typedef struct _DosRawTable		DosRawTable;
 
-/* note: this lots of bit-bashing here.  So you shouldn't look inside this.
+/* note: lots of bit-bashing here, thus, you shouldn't look inside it.
  * Use chs_to_sector() and sector_to_chs() instead.
  */
 typedef struct {
-	__u8	head;
-	__u8	sector;
-	__u8	cylinder;
+	uint8_t		head;
+	uint8_t		sector;
+	uint8_t		cylinder;
 } __attribute__((packed)) RawCHS;
 
 /* ripped from Linux source */
 struct _DosRawPartition {
-        __u8	boot_ind;	/* 00:  0x80 - active */
-	RawCHS	chs_start;	/* 01: */
-	__u8	type;		/* 04: partition type */
-	RawCHS	chs_end;	/* 05: */
-	__u32	start;		/* 08: starting sector counting from 0 */
-	__u32	length;		/* 0c: nr of sectors in partition */
+        uint8_t		boot_ind;	/* 00:  0x80 - active */
+	RawCHS		chs_start;	/* 01: */
+	uint8_t		type;		/* 04: partition type */
+	RawCHS		chs_end;	/* 05: */
+	uint32_t	start;		/* 08: starting sector counting from 0 */
+	uint32_t	length;		/* 0c: nr of sectors in partition */
 } __attribute__((packed));
 
 struct _DosRawTable {
 	char			boot_code [446];
 	DosRawPartition		partitions [4];
-	__u16			magic;
+	uint16_t			magic;
 } __attribute__((packed));
 
-void ped_disk_msdos_init ();
-void ped_disk_msdos_done ();
+typedef struct {
+	unsigned char	system;
+	int		boot;
+	int		hidden;
+	int		raid;
+	int		lvm;
+	int		lba;
+} DosPartitionData;
 
-static int msdos_probe (const PedDevice *dev);
-static PedDisk* msdos_open (PedDevice* dev);
-static PedDisk* msdos_create (PedDevice* dev);
-static int msdos_clobber (PedDevice* dev);
-static int msdos_close (PedDisk* disk);
-static int msdos_read (PedDisk* disk);
-static int msdos_write (PedDisk* disk);
-
-static PedPartition* msdos_partition_new (
-	const PedDisk* disk, PedPartitionType part_type,
-       	const PedFileSystemType* fs_type, PedSector start, PedSector end);
-static void msdos_partition_destroy (PedPartition* part);
-static int msdos_partition_set_flag (
-	PedPartition* part, PedPartitionFlag flag, int state);
-static int msdos_partition_get_flag (
-	const PedPartition* part, PedPartitionFlag flag);
-static int msdos_partition_is_flag_available (
-	const PedPartition* part,
-	PedPartitionFlag flag);
-static int msdos_partition_align (PedPartition* part,
-	       			  const PedConstraint* constraint);
-static int msdos_partition_enumerate (PedPartition* part);
-static int msdos_partition_set_extended_system (PedPartition* part);
-
-static int msdos_alloc_metadata (PedDisk* disk);
-static int msdos_get_max_primary_partition_count (const PedDisk* disk);
+static PedDiskType msdos_disk_type;
 
 static PedSector chs_to_sector (const PedDevice* dev, const RawCHS* chs);
 static int chs_to_cylinder (const RawCHS* chs);
-
-static PedDiskOps msdos_disk_ops = {
-	probe:			msdos_probe,
-	open:			msdos_open,
-	create:			msdos_create,
-	clobber:		msdos_clobber,
-	close:			msdos_close,
-	read:			msdos_read,
-	write:			msdos_write,
-
-	partition_new:		msdos_partition_new,
-	partition_destroy:	msdos_partition_destroy,
-	partition_set_flag:	msdos_partition_set_flag,
-	partition_get_flag:	msdos_partition_get_flag,
-	partition_is_flag_available:	msdos_partition_is_flag_available,
-	partition_set_name:	NULL,
-	partition_get_name:	NULL,
-	partition_align:	msdos_partition_align,
-	partition_enumerate:	msdos_partition_enumerate,
-	partition_set_extended_system:	msdos_partition_set_extended_system,
-
-	alloc_metadata:		msdos_alloc_metadata,
-	get_max_primary_partition_count:
-				msdos_get_max_primary_partition_count
-};
-
-static PedDiskType msdos_disk_type = {
-	next:		NULL,
-	name:		"msdos",
-	ops:		&msdos_disk_ops,
-	features:	PED_DISK_TYPE_EXTENDED
-};
-
-void
-ped_disk_msdos_init ()
-{
-	PED_ASSERT (sizeof (DosRawPartition) == 16, return);
-	PED_ASSERT (sizeof (DosRawTable) == 512, return);
-
-	ped_register_disk_type (&msdos_disk_type);
-}
-
-void
-ped_disk_msdos_done ()
-{
-	ped_unregister_disk_type (&msdos_disk_type);
-}
-
 static int raw_part_is_lba (const DosRawPartition* raw_part);
 
-/* sanity check for msdos_probe()
- *
- * We have to be very lenient, because the BIOS geometry may have been
- * detected incorrectly, or the size of the device, or something like that...
- *
- * We can't assert that legacy_start == raw_part->start :-/
- */
 static int
-check_partition_sanity (const PedDevice* dev, const DosRawPartition* raw_part)
-{
-	PedSector	legacy_start;
-	PedSector	legacy_end;
-	PedSector	max_chs_sector;
-	PedSector	cyl_size = dev->heads * dev->sectors;
-
-	if (raw_part->type == 0) {
-		DosRawPartition		empty;
-		memset (&empty, 0, sizeof (DosRawPartition));
-		return !memcmp (&empty, raw_part, sizeof (DosRawPartition));
-	}
-
-	if (memcmp (&raw_part->chs_start, "\0\0\0", 3) == 0)
-		return 0;
-	if (memcmp (&raw_part->chs_end, "\0\0\0", 3) == 0)
-		return 0;
-	if (!raw_part->start || !raw_part->length)
-		return 0;
-
-	/* somewhat subtle.  Windows ME generates broken partition tables.
-	 * Everyone agreed to have partitions > 1024 cylinder have the end
-	 * marked as cylinder 1023.  Windows ME, however, uses (end cyl % 1024)
-	 * However, these partitions are always marked LBA.  (A flag that
-	 * is only available for FAT partitions).  Hence, we should not do
-	 * the normal sanity checks, if the partition is marked as LBA.
-	 */
-	if (raw_part_is_lba (raw_part))
-		return 1;
-
-	legacy_start = chs_to_sector (dev, &raw_part->chs_start);
-	legacy_end = chs_to_sector (dev, &raw_part->chs_end);
-
-	max_chs_sector = (dev->sector_size / 512) * 1024 * 255 * 63;
-	if (legacy_start
-	    && PED_LE32_TO_CPU (raw_part->start) > max_chs_sector)
-		return 0;
-	if (legacy_end) {
-		if (PED_LE32_TO_CPU (raw_part->length) > max_chs_sector) {
-			/* subtle: some broken partition program might
-			 * have naively set end cylinder to > 1024, and
-			 * got wrapped around...
-			 */
-			if (chs_to_cylinder (&raw_part->chs_end)
-					== (legacy_end / cyl_size) % 1024)
-				return 1;
-			else 
-				return 0;
-		}
-		if (PED_LE32_TO_CPU (raw_part->start)
-		    + PED_LE32_TO_CPU (raw_part->length) - 1 > max_chs_sector)
-			return 0;
-	}
-
-	return 1;
-}
-
-static int
-msdos_probe (const PedDevice *dev)
+msdos_probe (PedDevice *dev)
 {
 	PedDiskType*	disk_type;
 	DosRawTable	part_table;
-	char		possible_pc98_table [PED_SECTOR_SIZE];
-	int		pc98_is_zero;
 	int		i;
 
 	PED_ASSERT (dev != NULL, return 0);
 
-	if (!ped_device_open ((PedDevice*) dev))
+	if (!ped_device_read (dev, &part_table, 0, 1))
 		return 0;
-	if (!ped_device_read (dev, &part_table, 0, 1)
-	    || !ped_device_read (dev, &possible_pc98_table, 1, 1)) {
-		ped_device_close ((PedDevice*) dev);
-		return 0;
-	}
-	ped_device_close ((PedDevice*) dev);
 
 	/* check magic */
 	if (PED_LE16_TO_CPU (part_table.magic) != MSDOS_MAGIC)
 		return 0;
 
-#ifdef ENABLE_PC98
-	pc98_is_zero = 1;
-	for (i = 0; i < PED_SECTOR_SIZE; i++) {
-		if (possible_pc98_table[i])
-			pc98_is_zero = 0;
-	}
-	if (!pc98_is_zero) {
-		/* check partition table sanity */
-		for (i = 0; i < 4; i++) {
-			if (!check_partition_sanity (dev,
-						     &part_table.partitions[i]))
-				return 0;
-		}
-	}
+	/* if this is a FAT fs, fail here */
+	if (!strncmp (part_table.boot_code + 0x36, "FAT", 3)
+	    || !strncmp (part_table.boot_code + 0x52, "FAT", 3))
+		return 0;
 
 	/* If this is a GPT disk, fail here */
 	for (i = 0; i < 4; i++) {
-		if (part_table.partitions[i].type == 0xEE)
- 			return 0;
- 	}
+		if (part_table.partitions[i].type == PARTITION_GPT)
+			return 0;
+	}
 
+#ifdef ENABLE_PC98
 	/* HACK: it's impossible to tell PC98 and msdos disk labels apart.
 	 * Someone made the signatures the same (very clever).  Since
 	 * PC98 has some idiosyncracies with it's boot-loader, it's detection
 	 * is more reliable */
-	disk_type = ped_disk_type_get (PC98_NAME);
+	disk_type = ped_disk_type_get ("pc98");
 	if (disk_type && disk_type->ops->probe (dev))
 		return 0;
 #endif /* ENABLE_PC98 */
@@ -285,78 +167,34 @@ msdos_probe (const PedDevice *dev)
 }
 
 static PedDisk*
-msdos_open (PedDevice* dev)
+msdos_alloc (PedDevice* dev)
 {
-	PedDisk*	disk;
+	PED_ASSERT (dev != NULL, return NULL);
 
-	PED_ASSERT (dev != NULL, return 0);
-
-	if (!msdos_probe (dev))
-		goto error;
-
-	if (!ped_device_open ((PedDevice*) dev))
-		goto error;
-	disk = ped_disk_alloc (dev, &msdos_disk_type);
-	if (!disk)
-		goto error_close_dev;
-
-	if (!msdos_read (disk))
-		goto error_free_disk;
-
-	return disk;
-
-error_free_disk:
-	ped_disk_free (disk);
-error_close_dev:
-	ped_device_close (dev);
-error:
-	return NULL;
+	return _ped_disk_alloc (dev, &msdos_disk_type);
 }
 
 static PedDisk*
-msdos_create (PedDevice* dev)
+msdos_duplicate (const PedDisk* disk)
 {
-	DosRawTable	table;
-
-	PED_ASSERT (dev != NULL, return 0);
-
-	if (!ped_device_open ((PedDevice*) dev))
-		goto error;
-	if (!ped_device_read (dev, &table, 0, 1))
-		goto error_close_dev;
-
-	if (!table.boot_code[0]) {
-		memset (table.boot_code, 0, 512);
-		memcpy (table.boot_code, MBR_BOOT_CODE,
-			sizeof (MBR_BOOT_CODE));
-	}
-
-	memset (table.partitions, 0, sizeof (DosRawPartition) * 4);
-	table.magic = PED_CPU_TO_LE16 (MSDOS_MAGIC);
-
-	if (!ped_device_write (dev, (void*) &table, 0, 1))
-		goto error_close_dev;
-	if (!ped_device_sync (dev))
-		goto error_close_dev;
-	ped_device_close (dev);
-	return msdos_open (dev);
-
-error_close_dev:
-	ped_device_close (dev);
-error:
-	return 0;
+	PedDisk*	new_disk;
+       
+	new_disk = ped_disk_new_fresh (disk->dev, &msdos_disk_type);
+	if (!new_disk)
+		return NULL;
+	new_disk->disk_specific = NULL;
+	return new_disk;
 }
 
-static int
-msdos_close (PedDisk* disk)
+static void
+msdos_free (PedDisk* disk)
 {
-	PED_ASSERT (disk != NULL, return 0);
+	PED_ASSERT (disk != NULL, return);
 
-	ped_device_close (disk->dev);
-	ped_disk_free (disk);
-	return 1;
+	_ped_disk_free (disk);
 }
 
+#ifndef DISCOVER_ONLY
 static int
 msdos_clobber (PedDevice* dev)
 {
@@ -365,25 +203,12 @@ msdos_clobber (PedDevice* dev)
 	PED_ASSERT (dev != NULL, return 0);
 	PED_ASSERT (msdos_probe (dev), return 0);
 
-	if (!ped_device_open ((PedDevice*) dev))
-		goto error;
 	if (!ped_device_read (dev, &table, 0, 1))
-		goto error_close_dev;
-
+		return 0;
 	table.magic = 0;
-
-	if (!ped_device_write (dev, (void*) &table, 0, 1))
-		goto error_close_dev;
-	if (!ped_device_sync (dev))
-		goto error_close_dev;
-	ped_device_close (dev);
-	return 1;
-
-error_close_dev:
-	ped_device_close (dev);
-error:
-	return 0;
+	return ped_device_write (dev, (void*) &table, 0, 1);
 }
+#endif /* !DISCOVER_ONLY */
 
 static PedSector
 chs_to_sector (const PedDevice* dev, const RawCHS* chs)
@@ -477,6 +302,7 @@ linear_end (const PedDisk* disk, const DosRawPartition* raw_part,
 	       	 	* (disk->dev->sector_size / 512);
 }
 
+#ifndef DISCOVER_ONLY
 static int
 probe_real_geom (PedDisk* disk, const DosRawPartition* raw_part,
 		 PedSector offset)
@@ -605,6 +431,7 @@ process_inconsistent_table (PedDisk* disk, const DosRawPartition* raw_part,
 	}
 	return 0;	/* should never get here - shut gcc up! */
 }
+#endif /* !DISCOVER_ONLY */
 
 static int
 raw_part_is_extended (DosRawPartition* raw_part)
@@ -612,9 +439,9 @@ raw_part_is_extended (DosRawPartition* raw_part)
 	PED_ASSERT (raw_part != NULL, return 0);
 
 	switch (raw_part->type) {
-	case 0x05:
-	case 0x0f:
-	case 0x85:
+	case PARTITION_EXT:
+	case PARTITION_EXT_LBA:
+	case PARTITION_LINUX_EXT:
 		return 1;
 
 	default:
@@ -630,12 +457,13 @@ raw_part_is_hidden (DosRawPartition* raw_part)
 	PED_ASSERT (raw_part != NULL, return 0);
 
 	switch (raw_part->type) {
-	case 0x11:
-	case 0x14:
-	case 0x16:
-	case 0x1b:
-	case 0x1c:
-	case 0x1e:
+	case PARTITION_FAT12_H:
+	case PARTITION_FAT16_SM_H:
+	case PARTITION_FAT16_H:
+	case PARTITION_FAT32_H:
+	case PARTITION_NTFS_H:
+	case PARTITION_FAT32_LBA_H:
+	case PARTITION_FAT16_LBA_H:
 		return 1;
 
 	default:
@@ -651,11 +479,11 @@ raw_part_is_lba (const DosRawPartition* raw_part)
 	PED_ASSERT (raw_part != NULL, return 0);
 
 	switch (raw_part->type) {
-	case 0xc:
-	case 0xe:
-	case 0xf:
-	case 0x1c:
-	case 0x1e:
+	case PARTITION_FAT32_LBA:
+	case PARTITION_FAT16_LBA:
+	case PARTITION_EXT_LBA:
+	case PARTITION_FAT32_LBA_H:
+	case PARTITION_FAT16_LBA_H:
 		return 1;
 
 	default:
@@ -665,6 +493,7 @@ raw_part_is_lba (const DosRawPartition* raw_part)
 	return 0;
 }
 
+#ifndef DISCOVER_ONLY
 static int
 check_consistancy (PedDisk* disk, DosRawPartition* raw_part, PedSector offset)
 {
@@ -685,6 +514,7 @@ check_consistancy (PedDisk* disk, DosRawPartition* raw_part, PedSector offset)
 		return process_inconsistent_table (disk, raw_part, offset);
 	return 1;
 }
+#endif /* !DISCOVER_ONLY */
 
 static int
 read_table (PedDisk* disk, PedSector sector, int is_extended_table)
@@ -710,6 +540,7 @@ read_table (PedDisk* disk, PedSector sector, int is_extended_table)
 	    && PED_LE16_TO_CPU (table.magic) == PARTITION_MAGIC_MAGIC)
 		return 1;
 
+#ifndef DISCOVER_ONLY
 	if (PED_LE16_TO_CPU (table.magic) != MSDOS_MAGIC) {
 		if (ped_exception_throw (
 			PED_EXCEPTION_ERROR, PED_EXCEPTION_IGNORE_CANCEL,
@@ -719,10 +550,11 @@ read_table (PedDisk* disk, PedSector sector, int is_extended_table)
 				!= PED_EXCEPTION_IGNORE)
 			goto error;
 	}
+#endif
 
 	for (i = 0; i < 4; i++) {
 		raw_part = &table.partitions [i];
-		if (raw_part->type == 0x0 || !raw_part->length)
+		if (raw_part->type == PARTITION_EMPTY || !raw_part->length)
 			continue;
 
 		/* process nested extended partitions after normal logical
@@ -736,8 +568,11 @@ read_table (PedDisk* disk, PedSector sector, int is_extended_table)
 		else
 			lba_offset = 0;
 
+#ifndef DISCOVER_ONLY
 		if (!check_consistancy (disk, raw_part, lba_offset))
 			return 0;
+#endif
+
 		part_start = linear_start (disk, raw_part, lba_offset);
 		if (part_start == sector) {
 			if (ped_exception_throw (
@@ -756,7 +591,7 @@ read_table (PedDisk* disk, PedSector sector, int is_extended_table)
 		else if (raw_part_is_extended (raw_part))
 			type = PED_PARTITION_EXTENDED;
 		else
-			type = PED_PARTITION_PRIMARY;
+			type = 0;
 
 		part = ped_partition_new (
 			disk, type, NULL,
@@ -768,9 +603,9 @@ read_table (PedDisk* disk, PedSector sector, int is_extended_table)
 		dos_data->system = raw_part->type;
 		dos_data->boot = raw_part->boot_ind != 0;
 		dos_data->hidden = raw_part_is_hidden (raw_part);
-		dos_data->raid = raw_part->type == 0xfd;
-		dos_data->lvm = raw_part->type == 0xfe
-					|| raw_part->type == 0x8e;
+		dos_data->raid = raw_part->type == PARTITION_LINUX_RAID;
+		dos_data->lvm = raw_part->type == PARTITION_LINUX_LVM_OLD
+				|| raw_part->type == PARTITION_LINUX_LVM;
 		dos_data->lba = raw_part_is_lba (raw_part);
 		if (!is_extended_table)
 			part->num = i + 1;
@@ -819,6 +654,7 @@ msdos_read (PedDisk* disk)
 	return read_table (disk, 0, 0);
 }
 
+#ifndef DISCOVER_ONLY
 static int
 fill_raw_part (DosRawPartition* raw_part, PedPartition* part, PedSector offset)
 {
@@ -831,14 +667,12 @@ fill_raw_part (DosRawPartition* raw_part, PedPartition* part, PedSector offset)
 	raw_part->boot_ind = 0x80 * dos_data->boot;
 	raw_part->type = dos_data->system;
 	raw_part->start = PED_CPU_TO_LE32 ((part->geom.start - offset)
-				/ (part->geom.disk->dev->sector_size / 512));
+				/ (part->disk->dev->sector_size / 512));
 	raw_part->length = PED_CPU_TO_LE32 (part->geom.length
-				/ (part->geom.disk->dev->sector_size / 512));
+				/ (part->disk->dev->sector_size / 512));
 
-	sector_to_chs (part->geom.disk->dev, part->geom.start,
-		       &raw_part->chs_start);
-	sector_to_chs (part->geom.disk->dev, part->geom.end,
-		       &raw_part->chs_end);
+	sector_to_chs (part->disk->dev, part->geom.start, &raw_part->chs_start);
+	sector_to_chs (part->disk->dev, part->geom.end, &raw_part->chs_end);
 
 	return 1;
 }
@@ -849,18 +683,17 @@ fill_ext_raw_part_geom (DosRawPartition* raw_part, PedGeometry* geom,
 {
 	PED_ASSERT (raw_part != NULL, return 0);
 	PED_ASSERT (geom != NULL, return 0);
-	PED_ASSERT (geom->disk != NULL, return 0);
-	PED_ASSERT (geom->disk->dev != NULL, return 0);
+	PED_ASSERT (geom->dev != NULL, return 0);
 
 	raw_part->boot_ind = 0;
-	raw_part->type = 0x5;
+	raw_part->type = PARTITION_EXT;
 	raw_part->start = PED_CPU_TO_LE32 ((geom->start - offset)
-				/ (geom->disk->dev->sector_size / 512));
+				/ (geom->dev->sector_size / 512));
 	raw_part->length = PED_CPU_TO_LE32 (geom->length
-				/ (geom->disk->dev->sector_size / 512));
+				/ (geom->dev->sector_size / 512));
 
-	sector_to_chs (geom->disk->dev, geom->start, &raw_part->chs_start);
-	sector_to_chs (geom->disk->dev, geom->start + geom->length - 1,
+	sector_to_chs (geom->dev, geom->start, &raw_part->chs_start);
+	sector_to_chs (geom->dev, geom->start + geom->length - 1,
 		       &raw_part->chs_end);
 
 	return 1;
@@ -888,7 +721,7 @@ write_ext_table (PedDisk* disk, PedSector sector, PedPartition* logical)
 
 	part = ped_disk_get_partition (disk, logical->num + 1);
 	if (part) {
-		geom = ped_geometry_new (disk, part->prev->geom.start,
+		geom = ped_geometry_new (disk->dev, part->prev->geom.start,
 				part->geom.end - part->prev->geom.start + 1);
 		if (!geom)
 			return 0;
@@ -970,10 +803,9 @@ msdos_write (PedDisk* disk)
 
 	if (!ped_device_write (disk->dev, (void*) &table, 0, 1))
 		return 0;
-	if (!ped_device_sync (disk->dev))
-		return 0;
-	return 1;
+	return ped_device_sync (disk->dev);
 }
+#endif /* !DISCOVER_ONLY */
 
 static PedPartition*
 msdos_partition_new (const PedDisk* disk, PedPartitionType part_type,
@@ -983,7 +815,7 @@ msdos_partition_new (const PedDisk* disk, PedPartitionType part_type,
 	PedPartition*		part;
 	DosPartitionData*	dos_data;
 
-	part = ped_partition_alloc (disk, part_type, fs_type, start, end);
+	part = _ped_partition_alloc (disk, part_type, fs_type, start, end);
 	if (!part)
 		goto error;
 
@@ -992,7 +824,7 @@ msdos_partition_new (const PedDisk* disk, PedPartitionType part_type,
 		       	= dos_data = ped_malloc (sizeof (DosPartitionData));
 		if (!dos_data)
 			goto error_free_part;
-		dos_data->system = 0x83;
+		dos_data->system = PARTITION_LINUX;
 		dos_data->hidden = 0;
 		dos_data->boot = 0;
 		dos_data->raid = 0;
@@ -1011,6 +843,29 @@ error:
 	return 0;
 }
 
+static PedPartition*
+msdos_partition_duplicate (const PedPartition* part)
+{
+	PedPartition*		new_part;
+	DosPartitionData*	new_dos_data;
+	DosPartitionData*	old_dos_data;
+
+	new_part = ped_partition_new (part->disk, part->type, part->fs_type,
+				      part->geom.start, part->geom.end);
+	if (!new_part)
+		return NULL;
+
+	old_dos_data = (DosPartitionData*) part->disk_specific;
+	new_dos_data = (DosPartitionData*) new_part->disk_specific;
+	new_dos_data->system = old_dos_data->system;
+	new_dos_data->boot = old_dos_data->boot;
+	new_dos_data->hidden = old_dos_data->hidden;
+	new_dos_data->raid = old_dos_data->raid;
+	new_dos_data->lvm = old_dos_data->lvm;
+	new_dos_data->lba = old_dos_data->lba;
+	return new_part;
+}
+
 static void
 msdos_partition_destroy (PedPartition* part)
 {
@@ -1021,24 +876,80 @@ msdos_partition_destroy (PedPartition* part)
 	ped_free (part);
 }
 
+static int
+msdos_partition_set_system (PedPartition* part,
+			    const PedFileSystemType* fs_type)
+{
+	DosPartitionData* dos_data = part->disk_specific;
+	PedSector cyl_size = part->disk->dev->sectors * part->disk->dev->heads;
+
+	part->fs_type = fs_type;
+
+	if (part->geom.end >= cyl_size * 1024
+	    && (part->type & PED_PARTITION_EXTENDED
+			|| (fs_type && !strncmp (fs_type->name, "fat", 3))))
+		dos_data->lba = 1;
+
+	if (dos_data->hidden
+		    && fs_type
+		    && strncmp (fs_type->name, "fat", 3) != 0
+		    && strcmp (fs_type->name, "ntfs") != 0)
+		dos_data->hidden = 0;
+
+	if (part->type & PED_PARTITION_EXTENDED) {
+		dos_data->raid = 0;
+		dos_data->lvm = 0;
+		if (dos_data->lba)
+			dos_data->system = PARTITION_EXT_LBA;
+		else
+			dos_data->system = PARTITION_EXT;
+		return 1;
+	}
+
+	if (dos_data->lvm) {
+		dos_data->system = PARTITION_LINUX_LVM;
+		return 1;
+	}
+	if (dos_data->raid) {
+		dos_data->system = PARTITION_LINUX_RAID;
+		return 1;
+	}
+
+	if (!fs_type)
+		dos_data->system = PARTITION_LINUX;
+	else if (!strcmp (fs_type->name, "fat16")) {
+		dos_data->system = dos_data->lba
+				   ? PARTITION_FAT16_LBA : PARTITION_FAT16;
+		dos_data->system |= dos_data->hidden ? PART_FLAG_HIDDEN : 0;
+	} else if (!strcmp (fs_type->name, "fat32")) {
+		dos_data->system = dos_data->lba
+				   ? PARTITION_FAT32_LBA : PARTITION_FAT32;
+		dos_data->system |= dos_data->hidden ? PART_FLAG_HIDDEN : 0;
+	} else if (!strcmp (fs_type->name, "ntfs")
+		   || !strcmp (fs_type->name, "hpfs")) {
+		dos_data->system = PARTITION_NTFS;
+		dos_data->system |= dos_data->hidden ? PART_FLAG_HIDDEN : 0;
+	} else if (!strcmp (fs_type->name, "linux-swap"))
+		dos_data->system = PARTITION_LINUX_SWAP;
+	else
+		dos_data->system = PARTITION_LINUX;
+
+	return 1;
+}
+
 static int
 msdos_partition_set_flag (PedPartition* part, PedPartitionFlag flag, int state)
 {
 	PedDisk*			disk;
 	PedPartition*			walk;
 	DosPartitionData*		dos_data;
-	const PedFileSystemType*	fs_type;
 
 	PED_ASSERT (part != NULL, return 0);
 	PED_ASSERT (part->disk_specific != NULL, return 0);
-	PED_ASSERT (part->geom.disk != NULL, return 0);
+	PED_ASSERT (part->disk != NULL, return 0);
 
 	dos_data = part->disk_specific;
-	disk = part->geom.disk;
-
-	/* hack! */
-	fs_type = part->fs_type ? part->fs_type
-				: ped_file_system_type_get ("ext2");
+	disk = part->disk;
 
 	switch (flag) {
 	case PED_PARTITION_HIDDEN:
@@ -1051,7 +962,7 @@ msdos_partition_set_flag (PedPartition* part, PedPartitionFlag flag, int state)
 			return 0;
 		}
 		dos_data->hidden = state;
-		return ped_partition_set_system (part, fs_type);
+		return ped_partition_set_system (part, part->fs_type);
 
 	case PED_PARTITION_BOOT:
 		dos_data->boot = state;
@@ -1072,8 +983,7 @@ msdos_partition_set_flag (PedPartition* part, PedPartitionFlag flag, int state)
 			dos_data->lvm = 0;
 		}
 		dos_data->raid = state;
-		dos_data->system = 0xfd;
-		return 1;
+		return ped_partition_set_system (part, part->fs_type);
 
 	case PED_PARTITION_LVM:
 		if (state) {
@@ -1081,12 +991,11 @@ msdos_partition_set_flag (PedPartition* part, PedPartitionFlag flag, int state)
 			dos_data->raid = 0;
 		}
 		dos_data->lvm = state;
-		dos_data->system = 0x8e;
-		return 1;
+		return ped_partition_set_system (part, part->fs_type);
 
 	case PED_PARTITION_LBA:
 		dos_data->lba = state;
-		return ped_partition_set_system (part, fs_type);
+		return ped_partition_set_system (part, part->fs_type);
 
 	default:
 		return 0;
@@ -1125,7 +1034,7 @@ msdos_partition_get_flag (const PedPartition* part, PedPartitionFlag flag)
 
 static int
 msdos_partition_is_flag_available (const PedPartition* part,
-	       			   PedPartitionFlag flag)
+				   PedPartitionFlag flag)
 {
 	switch (flag) {
 	case PED_PARTITION_HIDDEN:
@@ -1160,7 +1069,7 @@ _try_constraint (PedPartition* part, const PedConstraint* external,
 static PedGeometry*
 _best_solution (const PedPartition* part, PedGeometry* a, PedGeometry* b)
 {
-	PedDevice*	dev = part->geom.disk->dev;
+	PedDevice*	dev = part->disk->dev;
 	PedSector	cyl_size = dev->heads * dev->sectors;
 	int		a_cylinder;
 	int		b_cylinder;
@@ -1212,13 +1121,12 @@ choose_b:
 static PedConstraint*
 _primary_constraint (PedDisk* disk, PedGeometry* min_geom)
 {
+	PedDevice*	dev = disk->dev;
+	PedSector	cylinder_size = dev->sectors * dev->heads;
 	PedAlignment	start_align;
 	PedAlignment	end_align;
 	PedGeometry	start_geom;
 	PedGeometry	end_geom;
-	PedSector	cylinder_size;
-
-	cylinder_size = disk->dev->sectors * disk->dev->heads;
 
 	if (!ped_alignment_init (&start_align, 0, cylinder_size))
 		return NULL;
@@ -1228,22 +1136,22 @@ _primary_constraint (PedDisk* disk, PedGeometry* min_geom)
 	if (min_geom) {
 		if (min_geom->start < cylinder_size)
 			return NULL;
-		if (!ped_geometry_init (&start_geom, disk, cylinder_size,
+		if (!ped_geometry_init (&start_geom, dev, cylinder_size,
 			       		min_geom->start + 1 - cylinder_size))
 			return NULL;
-		if (!ped_geometry_init (&end_geom, disk, min_geom->end,
-			       		disk->dev->length - min_geom->end))
+		if (!ped_geometry_init (&end_geom, dev, min_geom->end,
+			       		dev->length - min_geom->end))
 			return NULL;
 	} else {
-		if (!ped_geometry_init (&start_geom, disk, cylinder_size,
-			       		disk->dev->length - cylinder_size))
+		if (!ped_geometry_init (&start_geom, dev, cylinder_size,
+			       		dev->length - cylinder_size))
 			return NULL;
-		if (!ped_geometry_init (&end_geom, disk, 0, disk->dev->length))
+		if (!ped_geometry_init (&end_geom, dev, 0, dev->length))
 			return NULL;
 	}
 
 	return ped_constraint_new (&start_align, &end_align, &start_geom,
-				   &end_geom, 1);
+				   &end_geom, 1, dev->length);
 }
 
 /* This constraint is for partitions starting on the first cylinder.  They
@@ -1252,35 +1160,33 @@ _primary_constraint (PedDisk* disk, PedGeometry* min_geom)
 static PedConstraint*
 _primary_start_constraint (PedDisk* disk, PedGeometry* min_geom)
 {
+	PedDevice*	dev = disk->dev;
+	PedSector	cylinder_size = dev->sectors * dev->heads;
 	PedAlignment	start_align;
 	PedAlignment	end_align;
 	PedGeometry	start_geom;
 	PedGeometry	end_geom;
-	PedSector	cylinder_size;
 
-	cylinder_size = disk->dev->sectors * disk->dev->heads;
-
-	if (!ped_alignment_init (&start_align, disk->dev->sectors, 0))
+	if (!ped_alignment_init (&start_align, dev->sectors, 0))
 		return NULL;
 	if (!ped_alignment_init (&end_align, -1, cylinder_size))
 		return NULL;
 	if (min_geom) {
-		if (!ped_geometry_init (&start_geom, disk, disk->dev->sectors,
-					1))
+		if (!ped_geometry_init (&start_geom, dev, dev->sectors, 1))
 			return NULL;
-		if (!ped_geometry_init (&end_geom, disk, min_geom->end,
-			       		disk->dev->length - min_geom->end))
+		if (!ped_geometry_init (&end_geom, dev, min_geom->end,
+			       		dev->length - min_geom->end))
 			return NULL;
 	} else {
-		if (!ped_geometry_init (&start_geom, disk, disk->dev->sectors,
-					disk->dev->length - disk->dev->sectors))
+		if (!ped_geometry_init (&start_geom, dev, dev->sectors,
+					dev->length - dev->sectors))
 			return NULL;
-		if (!ped_geometry_init (&end_geom, disk, 0, disk->dev->length))
+		if (!ped_geometry_init (&end_geom, dev, 0, dev->length))
 			return NULL;
 	}
 
 	return ped_constraint_new (&start_align, &end_align, &start_geom,
-				   &end_geom, 1);
+				   &end_geom, 1, dev->length);
 }
 
 /* constraints for logical partitions:
@@ -1295,7 +1201,8 @@ static PedConstraint*
 _logical_constraint (PedDisk* disk, PedSector start_offset, int is_start_part)
 {
 	PedPartition*	ext_part = ped_disk_extended_partition (disk);
-	PedSector	cylinder_size = disk->dev->sectors * disk->dev->heads;
+	PedDevice*	dev = disk->dev;
+	PedSector	cylinder_size = dev->sectors * dev->heads;
 	PedAlignment	start_align;
 	PedAlignment	end_align;
 	PedGeometry	max_geom;
@@ -1307,7 +1214,7 @@ _logical_constraint (PedDisk* disk, PedSector start_offset, int is_start_part)
 	if (!ped_alignment_init (&end_align, -1, cylinder_size))
 		return NULL;
 	if (is_start_part) {
-		if (!ped_geometry_init (&max_geom, disk,
+		if (!ped_geometry_init (&max_geom, dev,
 					ext_part->geom.start,
 					ext_part->geom.length))
 			return NULL;
@@ -1317,16 +1224,16 @@ _logical_constraint (PedDisk* disk, PedSector start_offset, int is_start_part)
 
 		min_start = ped_round_up_to (ext_part->geom.start + 1,
 					     cylinder_size);
+		max_length = ext_part->geom.end - min_start + 1;
 		if (min_start >= ext_part->geom.end)
 			return NULL;
 
-		max_length = ext_part->geom.end - min_start + 1;
-		if (!ped_geometry_init (&max_geom, disk, min_start, max_length))
+		if (!ped_geometry_init (&max_geom, dev, min_start, max_length))
 			return NULL;
 	}
 
 	return ped_constraint_new (&start_align, &end_align, &max_geom,
-		       		   &max_geom, 1);
+		       		   &max_geom, 1, dev->length);
 }
 
 /* returns the minimum geometry for the extended partition, given that the
@@ -1338,7 +1245,7 @@ _logical_constraint (PedDisk* disk, PedSector start_offset, int is_start_part)
 static PedGeometry*
 _get_min_extended_part_geom (const PedPartition* ext_part)
 {
-	PedDisk*		disk = ext_part->geom.disk;
+	PedDisk*		disk = ext_part->disk;
 	PedSector		head_size = disk->dev->sectors;
 	PedPartition*		walk;
 	PedGeometry*		min_geom;
@@ -1369,14 +1276,12 @@ _get_min_extended_part_geom (const PedPartition* ext_part)
 static int
 _align_primary (PedPartition* part, const PedConstraint* constraint)
 {
-	PedDisk*	disk;
-	PedSector	cylinder_size;
+	PedDisk*	disk = part->disk;
+	PedDevice*	dev = disk->dev;
+	PedSector	cylinder_size = dev->sectors * dev->heads;
 	PedGeometry*	min_geom = NULL;
 	PedGeometry*	solution = NULL;
 
-	disk = part->geom.disk;
-	cylinder_size = disk->dev->sectors * disk->dev->heads;
-
 	if (part->type == PED_PARTITION_EXTENDED)
 		min_geom = _get_min_extended_part_geom (part);
 
@@ -1405,7 +1310,7 @@ static int
 _logical_min_start_head (PedPartition* part, PedPartition* ext_part,
 			 int is_start_ext_part)
 {
-	PedDevice*	dev = part->geom.disk->dev;
+	PedDevice*	dev = part->disk->dev;
 	PedSector	cylinder_size = dev->sectors * dev->heads;
 	PedSector	base_head;
 
@@ -1424,7 +1329,7 @@ _logical_min_start_head (PedPartition* part, PedPartition* ext_part,
 static int
 _align_logical (PedPartition* part, const PedConstraint* constraint)
 {
-	PedDisk*	disk = part->geom.disk;
+	PedDisk*	disk = part->disk;
 	PedDevice*	dev = disk->dev;
 	PedPartition*	ext_part = ped_disk_extended_partition (disk);
 	PedSector	cyl_size = dev->sectors * dev->heads;
@@ -1475,41 +1380,39 @@ _align (PedPartition* part, const PedConstraint* constraint)
 static PedConstraint*
 _no_geom_constraint (PedDisk* disk, PedSector start, PedSector end)
 {
-	PedGeometry	 geom;
+	PedGeometry	 max;
 
-	ped_geometry_init (&geom, disk, start, end - start + 1);
-	return ped_constraint_new (ped_alignment_any, ped_alignment_any,
-			&geom, &geom, 1);
+	ped_geometry_init (&max, disk->dev, start, end - start + 1);
+	return ped_constraint_new_from_max (&max);
 }
 
 static PedConstraint*
 _no_geom_extended_constraint (PedPartition* part)
 {
-	PedDisk*	disk = part->geom.disk;
+	PedDevice*	dev = part->disk->dev;
 	PedGeometry*	min = _get_min_extended_part_geom (part);
 	PedGeometry	start_range;
 	PedGeometry	end_range;
 	PedConstraint*	constraint;
 
 	if (min) {
-		ped_geometry_init (&start_range, disk, 1, min->start);
-		ped_geometry_init (&end_range, disk, min->end,
-				   disk->dev->length - min->end);
+		ped_geometry_init (&start_range, dev, 1, min->start);
+		ped_geometry_init (&end_range, dev, min->end,
+				   dev->length - min->end);
 		ped_geometry_destroy (min);
 	} else {
-		ped_geometry_init (&start_range, disk, 1,
-				   disk->dev->length - 1);
-		ped_geometry_init (&end_range, disk, 1, disk->dev->length - 1);
+		ped_geometry_init (&start_range, dev, 1, dev->length - 1);
+		ped_geometry_init (&end_range, dev, 1, dev->length - 1);
 	}
 	constraint = ped_constraint_new (ped_alignment_any, ped_alignment_any,
-			&start_range, &end_range, 1);
+			&start_range, &end_range, 1, dev->length);
 	return constraint;
 }
 
 static int
 _align_primary_no_geom (PedPartition* part, const PedConstraint* constraint)
 {
-	PedDisk*	disk = part->geom.disk;
+	PedDisk*	disk = part->disk;
 	PedGeometry*	solution;
 
 	if (part->type == PED_PARTITION_EXTENDED) {
@@ -1533,7 +1436,7 @@ _align_primary_no_geom (PedPartition* part, const PedConstraint* constraint)
 static PedSector
 _calc_min_logical_start (PedPartition* part)
 {
-	PedDisk*	disk = part->geom.disk;
+	PedDisk*	disk = part->disk;
 	PedPartition*	ext_part = ped_disk_extended_partition (disk);
 	PedPartition*	walk;
 
@@ -1555,7 +1458,7 @@ _calc_min_logical_start (PedPartition* part)
 static int
 _align_logical_no_geom (PedPartition* part, const PedConstraint* constraint)
 {
-	PedDisk*	disk = part->geom.disk;
+	PedDisk*	disk = part->disk;
 	PedPartition*	ext_part = ped_disk_extended_partition (disk);
 	PedSector	min_start = _calc_min_logical_start (part);
 	PedGeometry*	solution;
@@ -1585,33 +1488,37 @@ _align_no_geom (PedPartition* part, const PedConstraint* constraint)
 static int
 msdos_partition_align (PedPartition* part, const PedConstraint* constraint)
 {
-	PED_ASSERT (part != NULL, return 0);
-
+ 	PED_ASSERT (part != NULL, return 0);
+ 
 	if (_align (part, constraint))
 		return 1;
-
-	if (part->geom.disk->dev->geom_known) {
-		if (ped_exception_throw (
-			PED_EXCEPTION_WARNING,
-			PED_EXCEPTION_IGNORE_CANCEL,
-			_("Unable to align partition properly.  This probably "
-			  "means that another partitioning tool generated an "
-			  "incorrect partition table, because it didn't have "
+ 
+	if (part->disk->dev->geom_known) {
+#ifndef DISCOVER_ONLY
+ 		if (ped_exception_throw (
+ 			PED_EXCEPTION_WARNING,
+ 			PED_EXCEPTION_IGNORE_CANCEL,
+ 			_("Unable to align partition properly.  This probably "
+ 			  "means that another partitioning tool generated an "
+ 			  "incorrect partition table, because it didn't have "
 			  "the correct BIOS geometry.  It is safe to ignore,"
 			  "but ignoring may cause (fixable) problems with "
 			  "some boot loaders."))
 				== PED_EXCEPTION_CANCEL)
-			return 0;
-		part->geom.disk->dev->geom_known = 0;
-	}
-
+ 			return 0;
+#endif /* !DISCOVER_ONLY */
+ 		part->disk->dev->geom_known = 0;
+ 	}
+ 
 	if (_align_no_geom (part, constraint))
 		return 1;
-
+ 
+#ifndef DISCOVER_ONLY
 	ped_exception_throw (
 		PED_EXCEPTION_ERROR,
 		PED_EXCEPTION_CANCEL,
-		_("Unable to align partition."));
+		_("Unable to satisfy all constraints on the partition."));
+#endif
 	return 0;
 }
 
@@ -1706,11 +1613,10 @@ msdos_alloc_metadata (PedDisk* disk)
 	cyl_size = dev->sectors * dev->heads;
 
 	if (dev->geom_known) {
-		if (!add_metadata_part (disk, PED_PARTITION_PRIMARY,
-					0, dev->sectors - 1))
+		if (!add_metadata_part (disk, 0, 0, dev->sectors - 1))
 			return 0;
 	} else {
-		if (!add_metadata_part (disk, PED_PARTITION_PRIMARY, 0, 0))
+		if (!add_metadata_part (disk, 0, 0, 0))
 			return 0;
 	}	
 
@@ -1718,7 +1624,7 @@ msdos_alloc_metadata (PedDisk* disk)
 	    && ped_round_down_to (dev->length, cyl_size) != dev->length) {
 		if (!add_metadata_part (
 				disk,
-				PED_PARTITION_PRIMARY,
+				0,
 				ped_round_down_to (dev->length, cyl_size),
 				dev->length - 1))
 			return 0;
@@ -1772,30 +1678,20 @@ next_logical (PedDisk* disk)
 static int
 msdos_partition_enumerate (PedPartition* part)
 {
-	PedDisk*	disk;
-
 	PED_ASSERT (part != NULL, return 0);
-	disk = part->geom.disk;
-	PED_ASSERT (disk != NULL, return 0);
+	PED_ASSERT (part->disk != NULL, return 0);
 
 	/* don't re-number a primary partition */
 	if (part->num != -1 && part->num <= 4)
 		return 1;
+
 	part->num = -1;
 
-	switch (part->type) {
-	case PED_PARTITION_PRIMARY:
-	case PED_PARTITION_EXTENDED:
-		part->num = next_primary (disk);
-		break;
+	if (part->type & PED_PARTITION_LOGICAL)
+		part->num = next_logical (part->disk);
+	else
+		part->num = next_primary (part->disk);
 
-	case PED_PARTITION_LOGICAL:
-		part->num = next_logical (disk);
-		break;
-
-	default:
-		return 0;
-	}
 	return 1;
 }
 
@@ -1805,33 +1701,59 @@ msdos_get_max_primary_partition_count (const PedDisk* disk)
 	return 4;
 }
 
-static int
-msdos_partition_set_extended_system (PedPartition* part)
+static PedDiskOps msdos_disk_ops = {
+	probe:			msdos_probe,
+#ifndef DISCOVER_ONLY
+	clobber:		msdos_clobber,
+#else
+	clobber:		NULL,
+#endif
+	alloc:			msdos_alloc,
+	duplicate:		msdos_duplicate,
+	free:			msdos_free,
+	read:			msdos_read,
+#ifndef DISCOVER_ONLY
+	write:			msdos_write,
+#else
+	write:			NULL,
+#endif
+
+	partition_new:		msdos_partition_new,
+	partition_duplicate:	msdos_partition_duplicate,
+	partition_destroy:	msdos_partition_destroy,
+	partition_set_system:	msdos_partition_set_system,
+	partition_set_flag:	msdos_partition_set_flag,
+	partition_get_flag:	msdos_partition_get_flag,
+	partition_is_flag_available:	msdos_partition_is_flag_available,
+	partition_set_name:	NULL,
+	partition_get_name:	NULL,
+	partition_align:	msdos_partition_align,
+	partition_enumerate:	msdos_partition_enumerate,
+
+	alloc_metadata:		msdos_alloc_metadata,
+	get_max_primary_partition_count:
+				msdos_get_max_primary_partition_count
+};
+
+static PedDiskType msdos_disk_type = {
+	next:		NULL,
+	name:		"msdos",
+	ops:		&msdos_disk_ops,
+	features:	PED_DISK_TYPE_EXTENDED
+};
+
+void
+ped_disk_msdos_init ()
 {
-	PedDisk*		disk;
-	DosPartitionData*	dos_data;
-	PedSector		cyl_size;
+	PED_ASSERT (sizeof (DosRawPartition) == 16, return);
+	PED_ASSERT (sizeof (DosRawTable) == 512, return);
 
-	PED_ASSERT (part != NULL, return 0);
-	PED_ASSERT (part->disk_specific != NULL, return 0);
+	ped_register_disk_type (&msdos_disk_type);
+}
 
-	disk = part->geom.disk;
-	cyl_size = disk->dev->sectors * disk->dev->heads;
-
-	dos_data = part->disk_specific;
-	dos_data->boot = 0;
-	dos_data->hidden = 0;
-	dos_data->raid = 0;
-	dos_data->lvm = 0;
-
-	if (part->geom.end / cyl_size >= 1024)
-		dos_data->lba = 1;
-
-	if (dos_data->lba)
-		dos_data->system = 0xf;
-	else
-		dos_data->system = 0x5;
-
-	return 1;
+void
+ped_disk_msdos_done ()
+{
+	ped_unregister_disk_type (&msdos_disk_type);
 }
 
