@@ -52,7 +52,6 @@
 #define MSDOS_MBR_SIGNATURE 0xaa55
 
 #define GPT_HEADER_SIGNATURE 0x5452415020494645
-#define GPT_HEADER_REVISION_V1_02 0x00010200
 #define GPT_HEADER_REVISION_V1_00 0x00010000
 #define GPT_HEADER_REVISION_V0_99 0x00009900
 
@@ -458,7 +457,7 @@ _parse_header (PedDisk* disk, GuidPartitionTableHeader_t* gpt)
 	PED_ASSERT (_header_is_valid (gpt), return 0);
 
 #ifndef DISCOVER_ONLY
-	if (PED_CPU_TO_LE32 (gpt->Revision) > GPT_HEADER_REVISION_V1_02
+	if (PED_CPU_TO_LE32 (gpt->Revision) > GPT_HEADER_REVISION_V1_00
 	    || PED_CPU_TO_LE32 (gpt->HeaderSize) != 92) {
 		if (ped_exception_throw (
 			PED_EXCEPTION_WARNING,
@@ -702,7 +701,7 @@ _generate_header (PedDisk* disk, int alternate, uint32_t ptes_crc,
 	memset (gpt, 0, sizeof (GuidPartitionTableHeader_t));
 
 	gpt->Signature = PED_CPU_TO_LE64 (GPT_HEADER_SIGNATURE);
-	gpt->Revision = PED_CPU_TO_LE32 (GPT_HEADER_REVISION_V1_02);
+	gpt->Revision = PED_CPU_TO_LE32 (GPT_HEADER_REVISION_V1_00);
 	gpt->HeaderSize = PED_CPU_TO_LE32 (92); /* per 1.02 spec */
 	gpt->HeaderCRC32 = 0;
 	gpt->Reserved1 = 0;
@@ -764,7 +763,7 @@ gpt_write(PedDisk * disk)
 	unsigned int i;
 
 	PED_ASSERT(disk != NULL, goto error);
-	PED_ASSERT(disk->dev != NULL, goto error)
+	PED_ASSERT(disk->dev != NULL, goto error);
 	PED_ASSERT(disk->disk_specific != NULL, goto error);
 
 	gpt_disk_data = disk->disk_specific;
