@@ -1,6 +1,6 @@
 /*
     libparted - a library for manipulating disk partitions
-    Copyright (C) 2000 Free Software Foundation, Inc.
+    Copyright (C) 2000, 2002 Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -20,12 +20,12 @@
 #include "config.h"
 
 #include <parted/parted.h>
+#include <parted/debug.h>
 #include <parted/endian.h>
-#include <parted/disk_mac.h>
 #include <string.h>
 
-#include <libintl.h>
 #if ENABLE_NLS
+#  include <libintl.h>
 #  define _(String) dgettext (PACKAGE, String)
 #else
 #  define _(String) (String)
@@ -80,135 +80,91 @@
 
 #define MAC_STATUS_BOOTABLE     8       /* partition is bootable */
 
-#define MIN_FREESPACE		(150 * 2)	/* 150k */
-
 typedef struct {
-	__u16   signature;      /* expected to be MAC_PARTITION_MAGIC */
-	__u16   res1;
-	__u32   map_count;      /* # blocks in partition map */
-	__u32   start_block;    /* absolute starting block # of partition */
-	__u32   block_count;    /* number of blocks in partition */
-	char    name[32];       /* partition name */
-	char    type[32];       /* string type description */
-	__u32   data_start;     /* rel block # of first data block */
-	__u32   data_count;     /* number of data blocks */
-	__u32   status;         /* partition status bits */
-	__u32   boot_start;
-	__u32   boot_count;
-	__u32   boot_load;
-	__u32   boot_load2;
-	__u32   boot_entry;
-	__u32   boot_entry2;
-	__u32   boot_cksum;
-	char    processor[16];  /* Contains 680x0, x=0,2,3,4; or empty */
-	__u32	driver_sig;
-	char	_padding[372];
+	uint16_t	signature;      /* expected to be MAC_PARTITION_MAGIC */
+	uint16_t	res1;
+	uint32_t	map_count;      /* # blocks in partition map */
+	uint32_t	start_block;    /* absolute starting block # of partition */
+	uint32_t	block_count;    /* number of blocks in partition */
+	char		name[32];       /* partition name */
+	char		type[32];       /* string type description */
+	uint32_t	data_start;     /* rel block # of first data block */
+	uint32_t	data_count;     /* number of data blocks */
+	uint32_t	status;         /* partition status bits */
+	uint32_t	boot_start;
+	uint32_t	boot_count;
+	uint32_t	boot_load;
+	uint32_t	boot_load2;
+	uint32_t	boot_entry;
+	uint32_t	boot_entry2;
+	uint32_t	boot_cksum;
+	char		processor[16];  /* Contains 680x0, x=0,2,3,4; or empty */
+	uint32_t	driver_sig;
+	char		_padding[372];
 } __attribute__ ((packed)) MacRawPartition;
 
 /* Driver descriptor structure, in block 0 */
 typedef struct {
-	__u16   signature;      /* expected to be MAC_DRIVER_MAGIC */
-	__u16   block_size;	/* usually 512  (occassionaly 2048?) */
-	__u32   block_count;	/* size of device in blocks */
-	__u16	dev_type;	/* unused? */
-	__u16	dev_id;		/* unused? */
-	__u32	data;		/* unused? */
-	__u16	driver_count;	/* # of driver descriptor entries */
-	__u32	start;		/* first driver's starting block */
-	__u16	size;		/* size of driver in blocks */
-	__u16	os_type;    	/* operating system type (MacOS = 1) */
-	char	reserved[486];	/* space reserved for the driver */
+	uint16_t	signature;      /* expected to be MAC_DRIVER_MAGIC */
+	uint16_t	block_size;	/* usually 512  (occassionaly 2048?) */
+	uint32_t	block_count;	/* size of device in blocks */
+	uint16_t	dev_type;	/* unused? */
+	uint16_t	dev_id;		/* unused? */
+	uint32_t	data;		/* unused? */
+	uint16_t	driver_count;	/* # of driver descriptor entries */
+	uint32_t	start;		/* first driver's starting block */
+	uint16_t	size;		/* size of driver in blocks */
+	uint16_t	os_type;    	/* operating system type (MacOS = 1) */
+	char		reserved[486];	/* space reserved for the driver */
 } __attribute__ ((packed)) MacRawDisk;
 
-void ped_disk_mac_init ();
-void ped_disk_mac_done ();
+typedef struct {
+	char		volume_name[33];	/* eg: "Games" */
+	char		system_name[33];	/* eg: "Apple_Unix_SVR2" */
+	char		processor_name[17];
 
-static int mac_probe (const PedDevice *dev);
-static PedDisk* mac_open (PedDevice* dev);
-static PedDisk* mac_create (PedDevice* dev);
-static int mac_clobber (PedDevice* dev);
-static int mac_close (PedDisk* disk);
-static int mac_read (PedDisk* disk);
-static int mac_write (PedDisk* disk);
+	int		is_boot;
+	int		is_driver;
+	int		is_root;
+	int		is_swap;
 
-static PedPartition* mac_partition_new (
-	const PedDisk* disk, PedPartitionType part_type,
-       	const PedFileSystemType* fs_type, PedSector start, PedSector end);
-static void mac_partition_destroy (PedPartition* part);
-static int mac_partition_set_flag (
-	PedPartition* part, PedPartitionFlag flag, int state);
-static int mac_partition_get_flag (
-	const PedPartition* part, PedPartitionFlag flag);
-static int mac_partition_is_flag_available (
-	const PedPartition* part,
-	PedPartitionFlag flag);
-static void mac_partition_set_name (PedPartition* part, const char* name);
-static const char* mac_partition_get_name (const PedPartition* part);
-static int mac_partition_align (PedPartition* part,
-				const PedConstraint* constraint);
-static int mac_partition_enumerate (PedPartition* part);
+	PedSector	data_region_length;
+	PedSector	boot_region_length;
 
-static int mac_alloc_metadata (PedDisk* disk);
-static int mac_get_max_primary_partition_count (const PedDisk* disk);
+	long long	boot_base_address;
+	long long	boot_entry_address;
+	int		boot_checksum;
 
+	uint32_t	status;
+	uint32_t	driver_sig;
+} MacPartitionData;
 
-static PedDiskOps mac_disk_ops = {
-	probe:			mac_probe,
-	open:			mac_open,
-	create:			mac_create,
-	clobber:		mac_clobber,
-	close:			mac_close,
-	read:			mac_read,
-	write:			mac_write,
-	
-	partition_new:		mac_partition_new,
-	partition_destroy:	mac_partition_destroy,
-	partition_set_flag:	mac_partition_set_flag,
-	partition_get_flag:	mac_partition_get_flag,
-	partition_is_flag_available:	mac_partition_is_flag_available,
-	partition_set_name:	mac_partition_set_name,
-	partition_get_name:	mac_partition_get_name,
-	partition_align:	mac_partition_align,
-	partition_enumerate:	mac_partition_enumerate,
-	partition_set_extended_system:	NULL,
+typedef struct {
+	int		ghost_size;		/* sectors per "driver" block */
+	int		part_map_entry_count;	/* # entries (incl. ghost) */
+	int		part_map_entry_num;	/* partition map location */
 
-	alloc_metadata:		mac_alloc_metadata,
-	get_max_primary_partition_count:
-				mac_get_max_primary_partition_count
-};
+	int		active_part_entry_count;	/* # real partitions */
+	int		free_part_entry_count;		/* # free space */
+	int		last_part_entry_num;		/* last entry number */
+} MacDiskData;
 
-static PedDiskType mac_disk_type = {
-	next:		NULL,
-	name:		MAC_NAME,
-	ops:		&mac_disk_ops,
-	features:	PED_DISK_TYPE_PARTITION_NAME
-};
-
-void
-ped_disk_mac_init ()
-{
-	PED_ASSERT (sizeof (MacRawPartition) == 512, return);
-	PED_ASSERT (sizeof (MacRawDisk) == 512, return);
-
-	ped_register_disk_type (&mac_disk_type);
-}
-
-void
-ped_disk_mac_done ()
-{
-	ped_unregister_disk_type (&mac_disk_type);
-}
+static PedDiskType mac_disk_type;
 
 static int
 _check_signature (MacRawDisk* raw_disk)
 {
 	if (PED_BE16_TO_CPU (raw_disk->signature) != MAC_DISK_MAGIC) {
+#ifdef DISCOVER_ONLY
+		return 0;
+#else
 		return ped_exception_throw (
 			PED_EXCEPTION_ERROR,
 			PED_EXCEPTION_IGNORE_CANCEL,
 			_("Invalid signature %x for Mac disk labels."),
 			(int) PED_BE16_TO_CPU (raw_disk->signature))
 			== PED_EXCEPTION_IGNORE;
+#endif
 	}
 
 	return 1;
@@ -222,88 +178,15 @@ _rawpart_check_signature (MacRawPartition* raw_part)
 }
 
 static int
-mac_probe (const PedDevice * dev)
+mac_probe (PedDevice * dev)
 {
 	MacRawDisk	buf;
 
 	PED_ASSERT (dev != NULL, return 0);
 
-	if (!ped_device_open ((PedDevice*) dev))
-		goto error;
-	if (!ped_device_read ((PedDevice*) dev, &buf, 0, 1))
-		goto error_close_dev;
-	ped_device_close ((PedDevice*) dev);
+	if (!ped_device_read (dev, &buf, 0, 1))
+		return 0;
 	return _check_signature (&buf);
-
-error_close_dev:
-	ped_device_close ((PedDevice*) dev);
-error:
-	return 0;
-}
-
-static PedDisk*
-mac_alloc (PedDevice* dev)
-{
-	PedDisk*		disk;
-
-	PED_ASSERT (dev != NULL, return NULL);
-
-	if (!ped_device_open ((PedDevice*) dev))
-		goto error;
-	if (dev->length < 256) {
-		ped_exception_throw (
-			PED_EXCEPTION_ERROR,
-			PED_EXCEPTION_CANCEL,
-			_("%s is too small for a Mac disk label!"),
-			dev->path);
-		goto error_close_dev;
-	}
-
-	disk = ped_disk_alloc (dev, &mac_disk_type);
-	if (!disk)
-		goto error_close_dev;
-
-	disk->disk_specific = (MacDiskData*) ped_malloc (sizeof (MacDiskData));
-	if (!disk->disk_specific)
-		goto error_free_disk;
-	return disk;
-
-error_free_disk:
-	ped_disk_free (disk);
-error_close_dev:
-	ped_device_close ((PedDevice*) dev);
-error:
-	return NULL;
-}
-
-static int
-mac_free (PedDisk* disk)
-{
-	MacDiskData*	mac_disk_data = disk->disk_specific;
-
-	ped_device_close (disk->dev);
-	ped_disk_free (disk);
-	ped_free (mac_disk_data);
-	return 1;
-}
-
-static PedDisk*
-mac_open (PedDevice* dev)
-{
-	PedDisk*		disk;
-
-	disk = mac_alloc (dev);
-	if (!disk)
-		goto error;
-	if (!mac_read (disk))
-		goto error_free_disk;
-
-	return disk;
-
-error_free_disk:
-	mac_free (disk);
-error:
-	return NULL;
 }
 
 static int
@@ -313,20 +196,21 @@ _disk_add_part_map_entry (PedDisk* disk, int warn)
 	PedPartition*		new_part;
 	MacPartitionData*	mac_part_data;
 	PedSector		part_map_size;
-	PedConstraint*		constraint_any = ped_constraint_any (disk);
+	PedConstraint*		constraint_any = ped_constraint_any (disk->dev);
 
+#ifndef DISCOVER_ONLY
 	if (warn && ped_exception_throw (
 		PED_EXCEPTION_ERROR,
 		PED_EXCEPTION_FIX | PED_EXCEPTION_CANCEL,
 		_("Partition map has no partition map entry!"))
 			!= PED_EXCEPTION_FIX)
 		goto error;
+#endif /* !DISCOVER_ONLY */
 
 	part_map_size
 		= ped_round_up_to (mac_disk_data->last_part_entry_num, 64);
 
-	new_part = ped_partition_new (disk, PED_PARTITION_PRIMARY, NULL,
-		       		      1, part_map_size - 1);
+	new_part = ped_partition_new (disk, 0, NULL, 1, part_map_size - 1);
 	if (!new_part)
 		goto error;
 
@@ -350,38 +234,81 @@ error:
 	return 0;
 }
 
-static PedDisk*
-mac_create (PedDevice* dev)
+PedDisk*
+mac_alloc (PedDevice* dev)
 {
 	PedDisk*		disk;
 	MacDiskData*		mac_disk_data;
 
-	disk = mac_alloc (dev);
+	PED_ASSERT (dev != NULL, return NULL);
+
+#ifndef DISCOVER_ONLY
+	if (dev->length < 256) {
+		ped_exception_throw (
+			PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_CANCEL,
+			_("%s is too small for a Mac disk label!"),
+			dev->path);
+		goto error;
+	}
+#endif
+
+	disk = _ped_disk_alloc (dev, &mac_disk_type);
 	if (!disk)
 		goto error;
 
-	mac_disk_data = disk->disk_specific;
+	mac_disk_data = (MacDiskData*) ped_malloc (sizeof (MacDiskData));
+	if (!disk->disk_specific)
+		goto error_free_disk;
+	disk->disk_specific = mac_disk_data;
 	mac_disk_data->ghost_size = disk->dev->sector_size / 512;
-
-	/* FIXME yuck - these should be set automagically somehow, rather
-	 * than the alloc_metadata mess
-	 */
 	mac_disk_data->active_part_entry_count = 0;
 	mac_disk_data->free_part_entry_count = 1;
 	mac_disk_data->last_part_entry_num = 1;
 
 	if (!_disk_add_part_map_entry (disk, 0))
 		goto error_free_disk;
-	if (!mac_write (disk))
-		goto error_free_disk;
 	return disk;
 
 error_free_disk:
-	mac_free (disk);
+	_ped_disk_free (disk);
 error:
 	return NULL;
 }
 
+static PedDisk*
+mac_duplicate (const PedDisk* disk)
+{
+	PedDisk*	new_disk;
+	MacDiskData*	new_mac_data;
+	MacDiskData*	old_mac_data = (MacDiskData*) disk->disk_specific;
+       
+	new_disk = ped_disk_new_fresh (disk->dev, &mac_disk_type);
+	if (!new_disk)
+		goto error;
+
+	new_mac_data = (MacDiskData*) new_disk->disk_specific;
+
+	/* ugly, but C is ugly :p */
+	memcpy (new_mac_data, old_mac_data, sizeof (MacDiskData));
+	return new_disk;
+
+error_free_new_disk:
+	_ped_disk_free (new_disk);
+error:
+	return NULL;
+}
+
+static void
+mac_free (PedDisk* disk)
+{
+	MacDiskData*	mac_disk_data = disk->disk_specific;
+
+	_ped_disk_free (disk);
+	ped_free (mac_disk_data);
+}
+
+#ifndef DISCOVER_ONLY
 static int
 _clobber_part_map (PedDevice* dev)
 {
@@ -404,36 +331,17 @@ mac_clobber (PedDevice* dev)
 {
 	MacRawDisk		raw_disk;
 
-	if (!ped_device_open (dev))
-		goto error;
-
 	if (!ped_device_read (dev, &raw_disk, 0, 1))
-		goto error_close_dev;
+		return 0;
 	if (!_check_signature (&raw_disk))
-		goto error_close_dev;
+		return 0;
 	memset (&raw_disk, 0, 512);
 	if (!ped_device_write (dev, &raw_disk, 0, 1))
-		goto error_close_dev;
+		return 0;
 
-	if (!_clobber_part_map (dev))
-		goto error_close_dev;
-
-	ped_device_close (dev);
-	return 1;
-
-error_close_dev:
-	ped_device_close (dev);
-error:
-	return 0;
-}
-
-static int
-mac_close (PedDisk* disk)
-{
-	PED_ASSERT (disk != NULL, return 0);
-
-	return mac_free (disk);
+	return _clobber_part_map (dev);
 }
+#endif /* !DISCOVER_ONLY */
 
 static int
 _rawpart_cmp_type (MacRawPartition* raw_part, char* type)
@@ -534,6 +442,7 @@ _rawpart_analyse (MacRawPartition* raw_part, PedDisk* disk, int num)
 	PedSector		start, length;
 
 	if (!_rawpart_check_signature (raw_part)) {
+#ifndef DISCOVER_ONLY
 		if (ped_exception_throw (
 			PED_EXCEPTION_WARNING,
 			PED_EXCEPTION_IGNORE_CANCEL,
@@ -541,6 +450,7 @@ _rawpart_analyse (MacRawPartition* raw_part, PedDisk* disk, int num)
 			num,
 			(int) PED_BE16_TO_CPU (raw_part->signature))
 				!= PED_EXCEPTION_IGNORE)
+#endif
 			goto error;
 	}
 
@@ -550,15 +460,16 @@ _rawpart_analyse (MacRawPartition* raw_part, PedDisk* disk, int num)
 	start = PED_BE32_TO_CPU (raw_part->start_block) * block_size;
 	length = PED_BE32_TO_CPU (raw_part->block_count) * block_size;
 	if (length == 0) {
+#ifndef DISCOVER_ONLY
 		ped_exception_throw (
 			PED_EXCEPTION_ERROR,
 			PED_EXCEPTION_CANCEL,
 			_("Partition %d has an invalid length of 0 bytes!"),
 			num);
+#endif
 		return NULL;
 	}
-	part = ped_partition_new (disk, PED_PARTITION_PRIMARY, NULL,
-				  start, start + length - 1);
+	part = ped_partition_new (disk, 0, NULL, start, start + length - 1);
 	if (!part)
 		goto error;
 
@@ -574,6 +485,7 @@ _rawpart_analyse (MacRawPartition* raw_part, PedDisk* disk, int num)
 	mac_part_data->is_swap = _rawpart_is_swap (raw_part);
 
 	/* "data" region */
+#ifndef DISCOVER_ONLY
 	if (raw_part->data_start) {
 		ped_exception_throw (
 			PED_EXCEPTION_ERROR,
@@ -582,10 +494,15 @@ _rawpart_analyse (MacRawPartition* raw_part, PedDisk* disk, int num)
 			  "of the partition"));
 		goto error_destroy_part;
 	}
+#endif /* !DISCOVER_ONLY */
 	mac_part_data->data_region_length
 		= PED_BE32_TO_CPU (raw_part->data_count) * block_size;
 
-	/* boot region */
+	/* boot region - we have no idea what this is for, but Mac OSX
+	 * seems to put garbage here, and doesn't pay any attention to
+	 * it afterwards.  [clausen, dan burcaw]
+	 */
+#if 0
 	if (raw_part->boot_start) {
 		ped_exception_throw (
 			PED_EXCEPTION_ERROR,
@@ -594,9 +511,11 @@ _rawpart_analyse (MacRawPartition* raw_part, PedDisk* disk, int num)
 			  "of the partition"));
 		goto error_destroy_part;
 	}
+#endif
 	mac_part_data->boot_region_length
 		= PED_BE32_TO_CPU (raw_part->boot_count) * block_size;
 
+#ifndef DISCOVER_ONLY
 	if (mac_part_data->is_driver) {
 		if (mac_part_data->boot_region_length < part->geom.length) {
 			if (ped_exception_throw (
@@ -618,6 +537,7 @@ _rawpart_analyse (MacRawPartition* raw_part, PedDisk* disk, int num)
 				goto error_destroy_part;
 		}
 	}
+#endif /* !DISCOVER_ONLY */
 
 	/* FIXME: should this add the high 32 bits (64 bit Macs?) */
 	mac_part_data->boot_base_address
@@ -661,17 +581,20 @@ _disk_analyse_block_size (PedDisk* disk, MacRawDisk* raw_disk)
 	PedSector	block_size;
 
 	if (PED_BE16_TO_CPU (raw_disk->block_size) % 512) {
+#ifndef DISCOVER_ONLY
 		ped_exception_throw (
 			PED_EXCEPTION_ERROR,
 			PED_EXCEPTION_CANCEL,
 			_("Weird block size on device descriptor: %d bytes is "
 			  "not divisible by 512."),
 			(int) PED_BE16_TO_CPU (raw_disk->block_size));
+#endif
 		goto error;
 	}
 
 	block_size = PED_BE16_TO_CPU (raw_disk->block_size) / 512;
 	if (block_size != disk->dev->sector_size / 512) {
+#ifndef DISCOVER_ONLY
 		if (ped_exception_throw (
 			PED_EXCEPTION_WARNING,
 			PED_EXCEPTION_IGNORE_CANCEL,
@@ -681,6 +604,7 @@ _disk_analyse_block_size (PedDisk* disk, MacRawDisk* raw_disk)
 			(int) disk->dev->sector_size)
 				!= PED_EXCEPTION_IGNORE)
 			goto error;
+#endif
 		disk->dev->sector_size = block_size * 512;
 	}
 
@@ -718,10 +642,12 @@ _disk_analyse_ghost_size (PedDisk* disk)
 		}
 	}
 
+#ifndef DISCOVER_ONLY
 	ped_exception_throw (
 		PED_EXCEPTION_ERROR,
 		PED_EXCEPTION_CANCEL,
 		_("No valid partition map found."));
+#endif
 	return 0;
 }
 
@@ -812,6 +738,7 @@ mac_read (PedDisk* disk)
 	if (!mac_disk_data->part_map_entry_num) {
 		if (!_disk_add_part_map_entry (disk, 1))
 			goto error_delete_all;
+		ped_disk_commit_to_dev (disk);
 	}
 	return 1;
 
@@ -821,6 +748,7 @@ error:
 	return 0;
 }
 
+#ifndef DISCOVER_ONLY
 /* The Ghost partition: is a blank entry, used to pad out each block (where
  * there physical block size > 512 bytes).  This is because OpenFirmware uses
  * 512 byte blocks, but device drivers Think Different TM, with a different
@@ -937,7 +865,7 @@ error:
 	return 0;	
 }
 
-int
+static int
 _generate_empty_part (PedDisk* disk, int num, MacRawPartition* part_map)
 {
 	MacDiskData*		mac_disk_data = disk->disk_specific;
@@ -991,7 +919,6 @@ static int
 mac_write (PedDisk* disk)
 {
 	MacRawPartition*	part_map;
-	PedDevice*		dev;
 	MacDiskData*		mac_disk_data;
 	PedPartition*		part;
 	int			num;
@@ -1002,7 +929,6 @@ mac_write (PedDisk* disk)
 	PED_ASSERT (!disk->update_mode, return 0);
 
 	mac_disk_data = disk->disk_specific;
-	dev = disk->dev;
 
 	if (!ped_disk_get_partition (disk, mac_disk_data->part_map_entry_num)) {
 		if (!_disk_add_part_map_entry (disk, 1))
@@ -1018,7 +944,7 @@ mac_write (PedDisk* disk)
 /* write (to memory) the "real" partitions */
 	for (part = ped_disk_next_partition (disk, NULL); part;
 	     part = ped_disk_next_partition (disk, part)) {
-		if (part->type != PED_PARTITION_PRIMARY)
+		if (!ped_partition_is_active (part))
 			continue;
 		if (!_generate_raw_part (disk, part, part_map))
 			goto error_free_part_map;
@@ -1041,21 +967,18 @@ mac_write (PedDisk* disk)
 		_generate_empty_part (disk, num, part_map);
 
 /* write to disk */
-	if (!ped_device_write (dev, part_map, 1,
+	if (!ped_device_write (disk->dev, part_map, 1,
 			       mac_disk_data->part_map_entry_count))
 		goto error_free_part_map;
 	ped_free (part_map);
-
-	if (!write_block_zero (disk))
-		goto error;
-
-	return 1;
+	return write_block_zero (disk);
 
 error_free_part_map:
 	ped_free (part_map);
 error:
 	return 0;
 }
+#endif /* !DISCOVER_ONLY */
 
 static PedPartition*
 mac_partition_new (
@@ -1065,7 +988,7 @@ mac_partition_new (
 	PedPartition*		part;
 	MacPartitionData*	mac_data;
 
-	part = ped_partition_alloc (disk, part_type, fs_type, start, end);
+	part = _ped_partition_alloc (disk, part_type, fs_type, start, end);
 	if (!part)
 		goto error;
 
@@ -1098,6 +1021,27 @@ error:
 	return 0;
 }
 
+static PedPartition*
+mac_partition_duplicate (const PedPartition* part)
+{
+	PedPartition*		new_part;
+	MacPartitionData*	new_mac_data;
+	MacPartitionData*	old_mac_data;
+
+	new_part = ped_partition_new (part->disk, part->type,
+				      part->fs_type, part->geom.start,
+				      part->geom.end);
+	if (!new_part)
+		return NULL;
+
+	old_mac_data = (MacPartitionData*) part->disk_specific;
+	new_mac_data = (MacPartitionData*) new_part->disk_specific;
+
+	/* ugly, but C is ugly :p */
+	memcpy (new_mac_data, old_mac_data, sizeof (MacPartitionData));
+	return new_part;
+}
+
 static void
 mac_partition_destroy (PedPartition* part)
 {
@@ -1108,77 +1052,75 @@ mac_partition_destroy (PedPartition* part)
 	ped_free (part);
 }
 
+static int
+mac_partition_set_system (PedPartition* part, const PedFileSystemType* fs_type)
+{
+	MacPartitionData* mac_data = part->disk_specific;
+
+	part->fs_type = fs_type;
+
+	if (fs_type && !strcmp (fs_type->name, "linux-swap"))
+		ped_partition_set_flag (part, PED_PARTITION_SWAP, 1);
+
+	if (mac_data->is_boot) {
+		strcpy (mac_data->system_name, "Apple_Bootstrap");
+		mac_data->status = 0x33;
+		return 1;
+	}
+
+	if (fs_type && !strcmp (fs_type->name, "hfs")) {
+		strcpy (mac_data->system_name, "Apple_HFS");
+		mac_data->status |= 0x7f;
+	} else {
+		strcpy (mac_data->system_name, "Apple_UNIX_SVR2");
+		mac_data->status = 0x33;
+	}
+
+	return 1;
+}
+
 static int
 mac_partition_set_flag (PedPartition* part, PedPartitionFlag flag, int state)
 {
 	PedFileSystemType*	hfs = ped_file_system_type_get ("hfs");
-	PedDisk*		disk;
 	MacPartitionData*	mac_data;
-	int			status;
 
 	PED_ASSERT (part != NULL, return 0);
 	PED_ASSERT (part->disk_specific != NULL, return 0);
-	PED_ASSERT (part->geom.disk != NULL, return 0);
 
 	mac_data = part->disk_specific;
-	disk = part->geom.disk;
 
 	switch (flag) {
 	case PED_PARTITION_BOOT:
 		mac_data->is_boot = state;
 
-		if (part->fs_type) {
-			status = ped_partition_set_system (part, part->fs_type);
-		} else {
-			strcpy (mac_data->system_name, "Apple_Bootstrap");
-			mac_data->status = 0x33;
-			status = 1;
-		}
-		return status;
-
-	case PED_PARTITION_ROOT:
-		if (!state) {
-			if (mac_data->is_root)
-				strcpy (mac_data->volume_name, "untitled");
-			mac_data->is_root = 0;
-			return 1;
-		}
-
-		if (strcasecmp (mac_data->system_name, "Apple_UNIX_SVR2")
-			    != 0) {
-			if (ped_exception_throw (
-				PED_EXCEPTION_ERROR,
-				PED_EXCEPTION_IGNORE_CANCEL,
-				_("Can't set non-Linux partitions as a root "
-				  "device."))
-					!= PED_EXCEPTION_IGNORE)
-				return 0;
-		}
-		mac_data->is_root = 1;
-		strcpy (mac_data->volume_name, "root");
+		if (part->fs_type)
+			return mac_partition_set_system (part, part->fs_type);
+			
+		strcpy (mac_data->system_name, "Apple_Bootstrap");
+		mac_data->status = 0x33;
 		return 1;
 
+	case PED_PARTITION_ROOT:
+		if (state) {
+			strcpy (mac_data->volume_name, "root");
+			mac_data->is_swap = 0;
+		} else {
+			if (mac_data->is_root)
+				strcpy (mac_data->volume_name, "untitled");
+		}
+		mac_data->is_root = state;
+		return 1;
 
 	case PED_PARTITION_SWAP:
-		if (!state) {
+		if (state) {
+			strcpy (mac_data->volume_name, "swap");
+			mac_data->is_root = 0;
+		} else {
 			if (mac_data->is_swap)
 				strcpy (mac_data->volume_name, "untitled");
-			mac_data->is_swap = 0;
-			return 1;
 		}
-
-		if (strcasecmp (mac_data->system_name, "Apple_UNIX_SVR2")
-			    != 0) {
-			if (ped_exception_throw (
-				PED_EXCEPTION_ERROR,
-				PED_EXCEPTION_IGNORE_CANCEL,
-				_("Can't set non-Linux partitions as a "
-				  "swap device."))
-					!= PED_EXCEPTION_IGNORE)
-				return 0;
-		}
-		mac_data->is_root = 1;
-		strcpy (mac_data->volume_name, "swap");
+		mac_data->is_swap = state;
 		return 1;
 
 	default:
@@ -1235,6 +1177,7 @@ mac_partition_set_name (PedPartition* part, const char* name)
 	PED_ASSERT (part->disk_specific != NULL, return);
 	mac_data = part->disk_specific;
 
+#ifndef DISCOVER_ONLY
 	if (mac_data->is_root || mac_data->is_swap) {
 		if (ped_exception_throw (
 			PED_EXCEPTION_WARNING,
@@ -1245,6 +1188,7 @@ mac_partition_set_name (PedPartition* part, const char* name)
 			return;
 		mac_data->is_root = mac_data->is_swap = 0;
 	}
+#endif
 
 	strncpy (mac_data->volume_name, name, 32);
 	mac_data->volume_name [32] = 0;
@@ -1279,110 +1223,11 @@ _primary_constraint (PedDisk* disk)
 		return NULL;
 	if (!ped_alignment_init (&end_align, -1, sector_size))
 		return NULL;
-	if (!ped_geometry_init (&max_geom, disk, 1, disk->dev->length - 1))
+	if (!ped_geometry_init (&max_geom, disk->dev, 1, disk->dev->length - 1))
 		return NULL;
 
 	return ped_constraint_new (&start_align, &end_align, &max_geom,
-				   &max_geom, 1);
-}
-
-/* argghh!  this should be somewhere else... but where?! */
-static int
-_try_constraint (PedPartition* part, const PedConstraint* external,
-		 PedConstraint* internal)
-{
-	PedConstraint*		intersection;
-	PedGeometry*		solution;
-
-	intersection = ped_constraint_intersect (external, internal);
-	ped_constraint_destroy (internal);
-	if (!intersection)
-		goto fail;
-
-	solution = ped_constraint_solve_nearest (intersection, &part->geom);
-	if (!solution)
-		goto fail_free_intersection;
-	ped_geometry_set (&part->geom, solution->start, solution->length);
-	ped_geometry_destroy (solution);
-	ped_constraint_destroy (intersection);
-	return 1;
-
-fail_free_intersection:
-	ped_constraint_destroy (intersection);
-fail:
-	return 0;
-}
-
-/* YUCK! FIXME (err, the whole new-then-add thing sucks) */
-static PedPartition*
-_find_left_part (PedPartition* part)
-{
-	PedPartition*	walk;
-
-	for (walk = part->geom.disk->part_list; walk && walk->next;
-	     walk = walk->next) {
-		if (walk->geom.end < part->geom.start
-		    && walk->next->geom.end >= part->geom.start)
-			break;
-	}
-	if (walk && walk->geom.end < part->geom.start)
-		return walk;
-	else
-		return NULL;
-}
-
-static PedPartition*
-_find_right_part (PedPartition* part, PedPartition* left)
-{
-	PedPartition*	walk = left;
-
-	for (walk = left; walk; walk = walk->next) {
-		if (walk->geom.start > part->geom.end)
-			break;
-	}
-	return walk;
-}
-
-/* if the gap between the new partition and the partition-to-be-aligned is
- * less than MIN_FREESPACE, then gobble up the gap!  (Only does this if it
- * still satisfies the constraint)
- */
-static int
-_grow_over_small_freespace (PedPartition* part, const PedConstraint* constraint)
-{
-	PedPartition*	left;
-	PedPartition*	right;
-
-	/* there should always be a partition on the left - if nothing else -
-	 * the partition map.  Unless "part" is the partition map ;-)
-	 */
-	left = _find_left_part (part);
-	if (!left)
-		return 1;
-	right = _find_right_part (part, left);
-	PED_ASSERT (left->geom.end < part->geom.start, return 0);
-	if (right)
-		PED_ASSERT (part->geom.end < right->geom.start, return 0);
-
-	if (part->geom.start - left->geom.end < MIN_FREESPACE
-	    && ped_alignment_is_aligned (constraint->start_align, NULL,
-					 left->geom.end + 1)
-	    && ped_geometry_test_sector_inside (constraint->start_range,
-		    				left->geom.end + 1)) {
-		if (!ped_geometry_set_start (&part->geom, left->geom.end + 1))
-			return 0;
-	}
-	if (right
-	    && right->geom.start - part->geom.end < MIN_FREESPACE
-	    && ped_alignment_is_aligned (constraint->end_align, NULL,
-					 right->geom.start - 1)
-	    && ped_geometry_test_sector_inside (constraint->end_range,
-		    				right->geom.start - 1)) {
-		if (!ped_geometry_set_end (&part->geom, right->geom.start - 1))
-			return 0;
-	}
-
-	return 1;
+				   &max_geom, 1, disk->dev->length);
 }
 
 static int
@@ -1390,19 +1235,17 @@ mac_partition_align (PedPartition* part, const PedConstraint* constraint)
 {
 	PED_ASSERT (part != NULL, return 0);
 
-	if (!_try_constraint (part, constraint,
-			     _primary_constraint (part->geom.disk))) {
-		ped_exception_throw (
-			PED_EXCEPTION_ERROR,
-			PED_EXCEPTION_CANCEL,
-			_("Unable to align partition."));
-		return 0;
-	}
+	if (_ped_partition_attempt_align (part, constraint,
+					  _primary_constraint (part->disk)))
+	       	return 1;
 
-	if (!_grow_over_small_freespace (part, constraint))
-		return 0;
-
-	return 1;
+#ifndef DISCOVER_ONLY
+	ped_exception_throw (
+		PED_EXCEPTION_ERROR,
+		PED_EXCEPTION_CANCEL,
+		_("Unable to satisfy all constraints on the partition."));
+#endif
+	return 0;
 }
 
 static int
@@ -1414,12 +1257,12 @@ mac_partition_enumerate (PedPartition* part)
 	int			max_part_count;
 
 	PED_ASSERT (part != NULL, return 0);
-	PED_ASSERT (part->geom.disk != NULL, return 0);
+	PED_ASSERT (part->disk != NULL, return 0);
 
-	disk = part->geom.disk;
+	disk = part->disk;
 	mac_disk_data = (MacDiskData*) disk->disk_specific;
 
-	max_part_count = mac_get_max_primary_partition_count (disk);
+	max_part_count = ped_disk_get_max_primary_partition_count (disk);
 
 	if (part->num > 0 && part->num <= mac_disk_data->part_map_entry_count)
 		return 1;
@@ -1431,10 +1274,12 @@ mac_partition_enumerate (PedPartition* part)
 		}
 	}
 
+#ifndef DISCOVER_ONLY
 	ped_exception_throw (
 		PED_EXCEPTION_ERROR,
 		PED_EXCEPTION_CANCEL,
 		_("Can't add another partition - partition map is too small!"));
+#endif
 
 	return 0;
 }
@@ -1483,14 +1328,12 @@ static int
 add_metadata_part (PedDisk* disk, PedSector start, PedSector end)
 {
 	PedPartition*		new_part;
-	PedConstraint*		constraint_any = ped_constraint_any (disk);
+	PedConstraint*		constraint_any = ped_constraint_any (disk->dev);
 
 	PED_ASSERT (disk != NULL, return 0);
 
-	new_part = ped_partition_new (
-			disk,
-		       	PED_PARTITION_PRIMARY | PED_PARTITION_METADATA, NULL,
-			start, end);
+	new_part = ped_partition_new (disk, PED_PARTITION_METADATA, NULL,
+				      start, end);
 	if (!new_part)
 		goto error;
 	if (!ped_disk_add_partition (disk, new_part, constraint_any))
@@ -1554,3 +1397,59 @@ mac_get_max_primary_partition_count (const PedDisk* disk)
 		- mac_disk_data->free_part_entry_count + 1;
 }
 
+static PedDiskOps mac_disk_ops = {
+	probe:			mac_probe,
+#ifndef DISCOVER_ONLY
+	clobber:		mac_clobber,
+#else
+	clobber:		NULL,
+#endif
+	alloc:			mac_alloc,
+	duplicate:		mac_duplicate,
+	free:			mac_free,
+	read:			mac_read,
+#ifndef DISCOVER_ONLY
+	write:			mac_write,
+#else
+	write:			NULL,
+#endif
+
+	partition_new:		mac_partition_new,
+	partition_duplicate:	mac_partition_duplicate,
+	partition_destroy:	mac_partition_destroy,
+	partition_set_system:	mac_partition_set_system,
+	partition_set_flag:	mac_partition_set_flag,
+	partition_get_flag:	mac_partition_get_flag,
+	partition_is_flag_available:	mac_partition_is_flag_available,
+	partition_set_name:	mac_partition_set_name,
+	partition_get_name:	mac_partition_get_name,
+	partition_align:	mac_partition_align,
+	partition_enumerate:	mac_partition_enumerate,
+
+	alloc_metadata:		mac_alloc_metadata,
+	get_max_primary_partition_count:
+				mac_get_max_primary_partition_count
+};
+
+static PedDiskType mac_disk_type = {
+	next:		NULL,
+	name:		"mac",
+	ops:		&mac_disk_ops,
+	features:	PED_DISK_TYPE_PARTITION_NAME
+};
+
+void
+ped_disk_mac_init ()
+{
+	PED_ASSERT (sizeof (MacRawPartition) == 512, return);
+	PED_ASSERT (sizeof (MacRawDisk) == 512, return);
+
+	ped_register_disk_type (&mac_disk_type);
+}
+
+void
+ped_disk_mac_done ()
+{
+	ped_unregister_disk_type (&mac_disk_type);
+}
+
