@@ -122,7 +122,6 @@ static char MBR_BOOT_CODE[] = {
 
 static PedDiskType pc98_disk_type;
 
-static int is_unused_partition(const PC98RawPartition* raw_part);
 static PedSector chs_to_sector (const PedDevice* dev, int c, int h, int s);
 static void sector_to_chs (const PedDevice* dev, PedSector sector,
 			   int* c, int* h, int* s);
@@ -157,6 +156,8 @@ check_partition_consistency (const PedDevice* dev,
 	   || PED_LE16_TO_CPU(raw_part->ipl_cyl) >= dev->cylinders
 	   || PED_LE16_TO_CPU(raw_part->cyl) >= dev->cylinders
 	   || PED_LE16_TO_CPU(raw_part->end_cyl) >= dev->cylinders
+	   || PED_LE16_TO_CPU(raw_part->cyl)
+	   	> PED_LE16_TO_CPU(raw_part->end_cyl)
 #if 0
 	   || !chs_to_sector(dev, PED_LE16_TO_CPU(raw_part->ipl_cyl),
 			     raw_part->ipl_head, raw_part->ipl_sect)
@@ -194,10 +195,9 @@ pc98_probe (PedDevice *dev)
 	     p < part_table.partitions + MAX_PART_COUNT;
 	     p++)
 	{
-		if (!is_unused_partition (p))
-			empty = 0;
 		if (p->mid == 0 && p->sid == 0)
 			continue;
+		empty = 0;
 		if (!check_partition_consistency (dev, p))
 			return 0;
 	}
