@@ -25,11 +25,11 @@
 #include <string.h>
 
 #include <parted/parted.h>
+#include <parted/debug.h>
 #include <parted/endian.h>
-#include <parted/disk_sun.h>
 
-#include <libintl.h>
 #if ENABLE_NLS
+#  include <libintl.h>
 #  define _(String) dgettext (PACKAGE, String)
 #else
 #  define _(String) (String)
@@ -77,76 +77,19 @@ typedef struct {
 	u_int16_t	csum;		/* Label xor'd checksum */
 }  __attribute__ ((packed)) SunRawLabel;
 
-static int sun_probe (const PedDevice *dev);
-static PedDisk* sun_open (PedDevice* dev);
-static PedDisk* sun_create (PedDevice* dev);
-static int sun_clobber (PedDevice* dev);
-static int sun_close (PedDisk* disk);
-static int sun_read (PedDisk* disk);
-static int sun_write (PedDisk* disk);
+typedef struct {
+	u_int8_t		type;
+	int			is_boot;
+	int			is_root;
+	int			is_lvm;
+} SunPartitionData;
 
-static PedPartition* sun_partition_new (
-	const PedDisk* disk, PedPartitionType part_type,
-       	const PedFileSystemType* fs_type, PedSector start, PedSector end);
-static void sun_partition_destroy (PedPartition* part);
-static int sun_partition_set_flag (
-	PedPartition* part, PedPartitionFlag flag, int state);
-static int sun_partition_get_flag (
-	const PedPartition* part, PedPartitionFlag flag);
-static int sun_partition_is_flag_available (
-	const PedPartition* part,
-	PedPartitionFlag flag);
-static int sun_partition_align (PedPartition* part,
-	       			  const PedConstraint* constraint);
-static int sun_partition_enumerate (PedPartition* part);
-static int sun_get_max_primary_partition_count (const PedDisk* disk);
+typedef struct {
+	PedSector		length; /* This is based on cyl - alt-cyl */
+	SunRawLabel		raw_label;
+} SunDiskData;
 
-static int sun_alloc_metadata (PedDisk* disk);
-
-static PedDiskOps sun_disk_ops = {
-	probe:			sun_probe,
-	open:			sun_open,
-	close:			sun_close,
-	read:			sun_read,
-	write:			sun_write,
-	create:			sun_create,
-	clobber:		sun_clobber,
-
-	partition_new:		sun_partition_new,
-	partition_destroy:	sun_partition_destroy,
-	partition_set_flag:	sun_partition_set_flag,
-	partition_get_flag:	sun_partition_get_flag,
-	partition_is_flag_available:	sun_partition_is_flag_available,
-	partition_align:	sun_partition_align,
-	partition_enumerate:	sun_partition_enumerate,
-	alloc_metadata:		sun_alloc_metadata,
-	get_max_primary_partition_count:
-				sun_get_max_primary_partition_count,
-
-	partition_set_name:		NULL,
-	partition_get_name:		NULL,
-	partition_set_extended_system:  NULL,
-};
-
-static PedDiskType sun_disk_type = {
-	next:		NULL,
-	name:		"sun",
-	ops:		&sun_disk_ops,
-	features:	0
-};
-
-void
-ped_disk_sun_init ()
-{
-	PED_ASSERT (sizeof (SunRawLabel) == 512, return);
-	ped_register_disk_type (&sun_disk_type);
-}
-
-void
-ped_disk_sun_done ()
-{
-	ped_unregister_disk_type (&sun_disk_type);
-}
+static PedDiskType sun_disk_type;
 
 /* Checksum computation */
 static void
@@ -173,43 +116,8 @@ sun_verify_checksum (SunRawLabel *label)
 	return !csum;
 }
 
-static PedDisk*
-sun_alloc (PedDevice* dev)
-{
-	PedDisk*                disk;
-
-	PED_ASSERT (dev != NULL, return NULL);
-
-	if (!ped_device_open ((PedDevice*) dev))
-		goto error;
-
-        disk = ped_disk_alloc (dev, &sun_disk_type);
-	if (!disk)
-		goto error_close_dev;
-
-	disk->disk_specific = (SunDiskData*) ped_malloc (sizeof (SunDiskData));
-	if (!disk->disk_specific)
-		goto error_free_disk;
-	return disk;
-
-error_free_disk:
-	ped_disk_free (disk);
-error_close_dev:
-	ped_device_close ((PedDevice*) dev);
-error:
-	return NULL;
-}
-
-static void
-sun_free (PedDisk *disk)
-{
-	ped_device_close (disk->dev);
-	ped_free (disk->disk_specific);
-	ped_disk_free (disk);
-}
-
 static int
-sun_probe (const PedDevice *dev)
+sun_probe (PedDevice *dev)
 {
 	PedDiskType*	disk_type;
 	SunRawLabel	label;
@@ -217,19 +125,14 @@ sun_probe (const PedDevice *dev)
 
 	PED_ASSERT (dev != NULL, return 0);
 
-	if (!ped_device_open ((PedDevice*) dev))
+	if (!ped_device_read (dev, &label, 0, 1))
 		return 0;
-	if (!ped_device_read (dev, &label, 0, 1)) {
-		ped_device_close ((PedDevice*) dev);
-		return 0;
-	}
-
-	ped_device_close ((PedDevice*) dev);
 
 	/* check magic */
 	if (PED_BE16_TO_CPU (label.magic) != SUN_DISK_MAGIC)
 		return 0;
 
+#ifndef DISCOVER_ONLY
 	if (!sun_verify_checksum(&label)) {
 		ped_exception_throw (
 			PED_EXCEPTION_ERROR,
@@ -237,89 +140,12 @@ sun_probe (const PedDevice *dev)
 			_("Corrupted Sun disk label detected."));
 		return 0;
 	}
+#endif
 
 	return 1;
 }
 
-static PedDisk*
-sun_open (PedDevice* dev)
-{
-	PedDisk*	disk;
-
-	disk = sun_alloc(dev);
-	if (!disk)
-		goto error;
-
-	if (!sun_read (disk))
-		goto error_free_disk;
-
-	return disk;
-
-error_free_disk:
-	sun_free (disk);
-error:
-	return NULL;
-}
-
-static int
-sun_close (PedDisk* disk)
-{
-	PED_ASSERT (disk != NULL, return 0);
-	sun_free(disk);
-	return 1;
-}
-
-static PedDisk*
-sun_create (PedDevice* dev)
-{
-	SunRawLabel	label;
-
-	PED_ASSERT (dev != NULL, return 0);
-
-	if (!ped_device_open ((PedDevice*) dev))
-		goto error;
-
-	memset(&label, 0, sizeof(SunRawLabel));
-
-	/* util-linux's fdisk asks about these...I'm just going to use
-	   defaults since most people do anyway...sue me.  */
-	label.magic	= PED_CPU_TO_BE16 (SUN_DISK_MAGIC);
-	label.nacyl	= PED_CPU_TO_BE16 (2);
-	label.pcylcount	= PED_CPU_TO_BE16 (dev->cylinders);
-	label.rspeed	= PED_CPU_TO_BE16 (5400);
-	label.ilfact	= PED_CPU_TO_BE16 (1);
-	label.sparecyl	= PED_CPU_TO_BE16 (dev->sectors);
-	label.ntrks	= PED_CPU_TO_BE16 (dev->heads);
-	label.nsect	= PED_CPU_TO_BE16 (dev->sectors);
-	label.ncyl	= PED_CPU_TO_BE16 (dev->cylinders - 2);
-
-	/* Add a whole disk partition at a minimum */
-	label.infos[WHOLE_DISK_PART].id = WHOLE_DISK_ID;
-	label.partitions[WHOLE_DISK_PART].start_cylinder = 0;
-	label.partitions[WHOLE_DISK_PART].num_sectors =
-		PED_CPU_TO_BE32((dev->cylinders-2) * dev->heads * dev->sectors);
-
-	/* Now a neato string to describe this label */
-	snprintf(label.info, sizeof(label.info) - 1,
-		 "GNU Parted Custom cyl %d alt 2 hd %d sec %d",
-		 PED_BE16_TO_CPU(label.ncyl), PED_BE16_TO_CPU(label.ntrks),
-		 PED_BE16_TO_CPU(label.nsect));
-
-	sun_compute_checksum(&label);
-
-	if (!ped_device_write (dev, &label, 0, 1))
-		goto error_close_dev;
-	if (!ped_device_sync (dev))
-		goto error_close_dev;
-	ped_device_close (dev);
-	return sun_open (dev);
-
-error_close_dev:
-	ped_device_close (dev);
-error:
-	return 0;
-}
-
+#ifndef DISCOVER_ONLY
 static int
 sun_clobber (PedDevice* dev)
 {
@@ -328,23 +154,86 @@ sun_clobber (PedDevice* dev)
 	PED_ASSERT (dev != NULL, return 0);
 	PED_ASSERT (sun_probe (dev), return 0);
 
-	if (!ped_device_open ((PedDevice*) dev))
-		goto error;
 	if (!ped_device_read (dev, &label, 0, 1))
-		goto error_close_dev;
+		return 0;
 	
 	label.magic = 0;
+	return ped_device_write (dev, &label, 0, 1);
+}
+#endif /* !DISCOVER_ONLY */
 
-	if (!ped_device_write (dev, &label, 0, 1))
-		goto error_close_dev;
-	
-	ped_device_close (dev);
-	return 1;
+static PedDisk*
+sun_alloc (PedDevice* dev)
+{
+	PedDisk*	disk;
+	SunRawLabel*	label;
 
-error_close_dev:
-	ped_device_close (dev);
+        disk = _ped_disk_alloc (dev, &sun_disk_type);
+	if (!disk)
+		goto error;
+
+	disk->disk_specific = (SunDiskData*) ped_malloc (sizeof (SunDiskData));
+	if (!disk->disk_specific)
+		goto error_free_disk;
+
+	label = &((SunDiskData*) disk->disk_specific)->raw_label;
+
+	memset(label, 0, sizeof(SunRawLabel));
+
+	/* util-linux's fdisk asks about these...I'm just going to use
+	   defaults since most people do anyway...sue me.  */
+	label->magic	= PED_CPU_TO_BE16 (SUN_DISK_MAGIC);
+	label->nacyl	= PED_CPU_TO_BE16 (2);
+	label->pcylcount	= PED_CPU_TO_BE16 (dev->cylinders);
+	label->rspeed	= PED_CPU_TO_BE16 (5400);
+	label->ilfact	= PED_CPU_TO_BE16 (1);
+	label->sparecyl	= PED_CPU_TO_BE16 (dev->sectors);
+	label->ntrks	= PED_CPU_TO_BE16 (dev->heads);
+	label->nsect	= PED_CPU_TO_BE16 (dev->sectors);
+	label->ncyl	= PED_CPU_TO_BE16 (dev->cylinders - 2);
+
+	/* Add a whole disk partition at a minimum */
+	label->infos[WHOLE_DISK_PART].id = WHOLE_DISK_ID;
+	label->partitions[WHOLE_DISK_PART].start_cylinder = 0;
+	label->partitions[WHOLE_DISK_PART].num_sectors =
+		PED_CPU_TO_BE32((dev->cylinders-2) * dev->heads * dev->sectors);
+
+	/* Now a neato string to describe this label */
+	snprintf(label->info, sizeof(label->info) - 1,
+		 "GNU Parted Custom cyl %d alt 2 hd %d sec %d",
+		 PED_BE16_TO_CPU(label->ncyl), PED_BE16_TO_CPU(label->ntrks),
+		 PED_BE16_TO_CPU(label->nsect));
+
+	sun_compute_checksum(label);
+	return disk;
+
+error_free_disk:
+	_ped_disk_free (disk);
 error:
-	return 0;
+	return NULL;
+}
+
+static PedDisk*
+sun_duplicate (const PedDisk* disk)
+{
+	PedDisk*	new_disk;
+	SunDiskData*	new_sun_data;
+	SunDiskData*	old_sun_data = (SunDiskData*) disk->disk_specific;
+       
+	new_disk = ped_disk_new_fresh (disk->dev, &sun_disk_type);
+	if (!new_disk)
+		return NULL;
+
+	new_sun_data = (SunDiskData*) new_disk->disk_specific;
+	memcpy (new_sun_data, old_sun_data, sizeof (SunDiskData));
+	return new_disk;
+}
+
+static void
+sun_free (PedDisk *disk)
+{
+	ped_free (disk->disk_specific);
+	_ped_disk_free (disk);
 }
 
 static int
@@ -354,6 +243,7 @@ _check_geometry_sanity (PedDisk* disk, SunRawLabel* label)
 
 	if (PED_BE16_TO_CPU(label->nsect) != dev->sectors ||
 	    PED_BE16_TO_CPU(label->ntrks) != dev->heads) {
+#ifndef DISCOVER_ONLY
 		if (ped_exception_throw (
 				PED_EXCEPTION_WARNING,
 				PED_EXCEPTION_IGNORE_CANCEL,
@@ -366,6 +256,7 @@ _check_geometry_sanity (PedDisk* disk, SunRawLabel* label)
 				PED_BE16_TO_CPU(label->nsect))
 			== PED_EXCEPTION_CANCEL)
 			return 0;
+#endif
 		dev->sectors = PED_BE16_TO_CPU(label->nsect);
 		dev->heads = PED_BE16_TO_CPU(label->ntrks);
 		dev->cylinders = PED_BE16_TO_CPU(label->pcylcount);
@@ -387,7 +278,7 @@ _check_geometry_sanity (PedDisk* disk, SunRawLabel* label)
 static int
 sun_read (PedDisk* disk)
 {
-	SunRawLabel label;
+	SunRawLabel* label;
 	SunPartitionData* sun_data;
 	SunDiskData* disk_data;
 	int i;
@@ -399,41 +290,39 @@ sun_read (PedDisk* disk)
 	PED_ASSERT (disk->dev != NULL, return 0);
 	PED_ASSERT (disk->disk_specific != NULL, return 0);
 	
+	disk_data = (SunDiskData*) disk->disk_specific;
+	label = &disk_data->raw_label;
+
 	ped_disk_delete_all (disk);
 
-	if (!ped_device_open ((PedDevice*) disk->dev))
+	if (!ped_device_read (disk->dev, label, 0, 1))
+		goto error;
+	if (!_check_geometry_sanity (disk, label))
 		goto error;
-	if (!ped_device_read (disk->dev, &label, 0, 1))
-		goto error_close_dev;
 
-	if (!_check_geometry_sanity (disk, &label))
-		goto error_close_dev;
-
-	disk_data = disk->disk_specific;
 	block = disk->dev->sectors * disk->dev->heads;
 	disk_data->length = block * (disk->dev->cylinders
-					- PED_BE16_TO_CPU(label.nacyl));
+					- PED_BE16_TO_CPU(label->nacyl));
 
 	for (i = 0; i < SUN_DISK_MAXPARTITIONS; i++) {
-		if (!PED_BE32_TO_CPU(label.partitions[i].num_sectors))
+		if (!PED_BE32_TO_CPU(label->partitions[i].num_sectors))
 			continue;
-		if (!label.infos[i].id)
+		if (!label->infos[i].id)
 			continue;
-		if (label.infos[i].id == WHOLE_DISK_ID)
+		if (label->infos[i].id == WHOLE_DISK_ID)
 	       		continue;
 
-		start = PED_BE32_TO_CPU(label.partitions[i].start_cylinder)
+		start = PED_BE32_TO_CPU(label->partitions[i].start_cylinder)
 				    * block;
 		end = start
-		      + PED_BE32_TO_CPU(label.partitions[i].num_sectors) - 1;
+		      + PED_BE32_TO_CPU(label->partitions[i].num_sectors) - 1;
 
-		part = ped_partition_new (disk, PED_PARTITION_PRIMARY, NULL,
-				  start, end);
+		part = ped_partition_new (disk, 0, NULL, start, end);
 		if (!part)
-			goto error_close_dev;
+			goto error;
 
 		sun_data = part->disk_specific;
-		sun_data->type = label.infos[i].id;
+		sun_data->type = label->infos[i].id;
 		sun_data->is_boot = sun_data->type == 0x1;
 		sun_data->is_root = sun_data->type == 0x2;
 		sun_data->is_lvm = sun_data->type == 0x8e;
@@ -443,23 +332,36 @@ sun_read (PedDisk* disk)
 	
 		constraint_exact = ped_constraint_exact (&part->geom);
 		if (!ped_disk_add_partition (disk, part, constraint_exact))
-			goto error_close_dev;
+			goto error;
 		ped_constraint_destroy (constraint_exact);
 	}
 
-	ped_device_close ((PedDevice*) disk->dev);
 	return 1;
 
- error_close_dev:
-	ped_device_close ((PedDevice*) disk->dev);
  error:
 	return 0;
 }
 
+#ifndef DISCOVER_ONLY
+static void
+_probe_and_add_use_old_info (PedDisk* disk)
+{
+	SunDiskData*		sun_specific;
+	SunRawLabel		old_label;
+
+	sun_specific = (SunDiskData*) disk->disk_specific;
+
+	if (!ped_device_read (disk->dev, &old_label, 0, 1))
+		return;
+	if (old_label.info [0]
+	    && old_label.magic == PED_CPU_TO_LE32 (SUN_DISK_MAGIC))
+		memcpy (&sun_specific->raw_label, &old_label, 512);
+}
+
 static int
 sun_write (PedDisk* disk)
 {
-	SunRawLabel		label;
+	SunRawLabel*		label;
 	SunPartitionData*	sun_data;
 	SunDiskData*		disk_data;
 	PedPartition*		part;
@@ -469,16 +371,16 @@ sun_write (PedDisk* disk)
 	PED_ASSERT (disk != NULL, return 0);
 	PED_ASSERT (disk->dev != NULL, return 0);
 
-	if (!ped_device_read (disk->dev, &label, 0, 1))
-		return 0;
+	_probe_and_add_use_old_info (disk);
 
-	memset (label.partitions, 0,
+	disk_data = (SunDiskData*) disk->disk_specific;
+	label = &disk_data->raw_label;
+
+	memset (label->partitions, 0,
 		sizeof (SunRawPartition) * SUN_DISK_MAXPARTITIONS);
-	memset (label.infos, 0,
+	memset (label->infos, 0,
 		sizeof (SunPartitionInfo) * SUN_DISK_MAXPARTITIONS);
 
-	disk_data = disk->disk_specific;
-
 	for (i = 0; i < SUN_DISK_MAXPARTITIONS; i++) {
 		part = ped_disk_get_partition (disk, i + 1);
 
@@ -487,9 +389,9 @@ sun_write (PedDisk* disk)
 			   partition, so let's put it there for safety
 			   sake.  */
 
-			label.infos[i].id = WHOLE_DISK_ID;
-			label.partitions[i].start_cylinder = 0;
-			label.partitions[i].num_sectors = 
+			label->infos[i].id = WHOLE_DISK_ID;
+			label->partitions[i].start_cylinder = 0;
+			label->partitions[i].num_sectors = 
 				PED_CPU_TO_BE32(disk_data->length);
 			continue;
 		}
@@ -497,11 +399,11 @@ sun_write (PedDisk* disk)
 			continue;
 
 		sun_data = part->disk_specific;
-		label.infos[i].id = sun_data->type;
-		label.partitions[i].start_cylinder
+		label->infos[i].id = sun_data->type;
+		label->partitions[i].start_cylinder
 			= PED_CPU_TO_BE32 (part->geom.start
 				/ (disk->dev->sectors * disk->dev->heads));
-		label.partitions[i].num_sectors
+		label->partitions[i].num_sectors
 			= PED_CPU_TO_BE32 (part->geom.end
 					   - part->geom.start + 1);
 	}
@@ -514,19 +416,20 @@ sun_write (PedDisk* disk)
 	   smaller than it really is, but we'll have that problem even if we
 	   don't do this.  */
 
-	label.pcylcount = PED_CPU_TO_BE16 (disk->dev->cylinders);
-	label.ncyl = PED_CPU_TO_BE16 (disk->dev->cylinders
-	       				- PED_BE32_TO_CPU(label.nacyl));
+	label->pcylcount = PED_CPU_TO_BE16 (disk->dev->cylinders);
+	label->ncyl = PED_CPU_TO_BE16 (disk->dev->cylinders
+	       				- PED_BE32_TO_CPU(label->nacyl));
 
-	sun_compute_checksum(&label);
+	sun_compute_checksum (label);
 
-	if (!ped_device_write (disk->dev, &label, 0, 1))
-		return 0;
-	if (!ped_device_sync (disk->dev))
-		return 0;
+	if (!ped_device_write (disk->dev, label, 0, 1))
+		goto error;
+	return ped_device_sync (disk->dev);
 
-	return 1;
+error:
+	return 0;
 }
+#endif /* !DISCOVER_ONLY */
 
 static PedPartition*
 sun_partition_new (const PedDisk* disk, PedPartitionType part_type,
@@ -536,7 +439,7 @@ sun_partition_new (const PedDisk* disk, PedPartitionType part_type,
 	PedPartition*		part;
 	SunPartitionData*	sun_data;
 
-	part = ped_partition_alloc (disk, part_type, fs_type, start, end);
+	part = _ped_partition_alloc (disk, part_type, fs_type, start, end);
 	if (!part)
 		goto error;
 
@@ -563,6 +466,27 @@ error:
 	return NULL;
 }
 
+static PedPartition*
+sun_partition_duplicate (const PedPartition* part)
+{
+	PedPartition*		new_part;
+	SunPartitionData*	new_sun_data;
+	SunPartitionData*	old_sun_data;
+
+	new_part = ped_partition_new (part->disk, part->type,
+				      part->fs_type, part->geom.start,
+				      part->geom.end);
+	if (!new_part)
+		return NULL;
+
+	old_sun_data = (SunPartitionData*) part->disk_specific;
+	new_sun_data = (SunPartitionData*) new_part->disk_specific;
+	new_sun_data->type = old_sun_data->type;
+	new_sun_data->is_boot = old_sun_data->is_boot;
+	new_sun_data->is_root = old_sun_data->is_root;
+	new_sun_data->is_lvm = old_sun_data->is_lvm;
+	return new_part;
+}
 
 static void
 sun_partition_destroy (PedPartition* part)
@@ -574,43 +498,66 @@ sun_partition_destroy (PedPartition* part)
 	ped_free (part);
 }
 
+static int
+sun_partition_set_system (PedPartition* part, const PedFileSystemType* fs_type)
+{
+	SunPartitionData*		sun_data = part->disk_specific;
+
+	part->fs_type = fs_type;
+
+	if (sun_data->is_boot) {
+		sun_data->type = 0x1;
+		return 1;
+	}
+	if (sun_data->is_root) {
+		sun_data->type = 0x2;
+		return 1;
+	}
+	if (sun_data->is_lvm) {
+		sun_data->type = 0x8e;
+		return 1;
+	}
+
+	sun_data->type = 0x83;
+	if (fs_type) {
+		if (!strcmp (fs_type->name, "linux-swap"))
+			sun_data->type = 0x82;
+		else if (!strcmp (fs_type->name, "ufs"))
+			sun_data->type = 0x6;
+	}
+
+	return 1;
+}
 
 static int
 sun_partition_set_flag (PedPartition* part, PedPartitionFlag flag, int state)
 {
-	SunPartitionData*	       sun_data;
-	const PedFileSystemType*	fs_type;
+	SunPartitionData*		sun_data;
 
 	PED_ASSERT (part != NULL, return 0);
 	PED_ASSERT (part->disk_specific != NULL, return 0);
-	PED_ASSERT (sun_partition_is_flag_available (part, flag), return 0);
+	PED_ASSERT (ped_partition_is_flag_available (part, flag), return 0);
 
 	sun_data = part->disk_specific;
 
-	/* hack! */
-	fs_type = part->fs_type ? part->fs_type
-				: ped_file_system_type_get ("ext2");
-
 	switch (flag) {
 		case PED_PARTITION_BOOT:
 			sun_data->is_boot = state;
 			if (state)
 				sun_data->is_root = sun_data->is_lvm = 0;
-			sun_data->type = 0x1;
-			return 1;
+			return ped_partition_set_system (part, part->fs_type);
 
 		case PED_PARTITION_ROOT:
 			sun_data->is_root = state;
 			if (state)
 				sun_data->is_boot = sun_data->is_lvm = 0;
-			return ped_partition_set_system (part, fs_type);
+			return ped_partition_set_system (part, part->fs_type);
 
 		case PED_PARTITION_LVM:
 			sun_data->is_lvm = state;
 			if (state)
 				sun_data->is_root = sun_data->is_boot = 0;
-			sun_data->type = 0x8e;
-			return 1;
+			return ped_partition_set_system (part, part->fs_type);
 
 		default:
 			return 0;
@@ -667,65 +614,40 @@ sun_get_max_primary_partition_count (const PedDisk* disk)
 static PedConstraint*
 _get_strict_constraint (PedDisk* disk)
 {
+	PedDevice*	dev = disk->dev;
         PedAlignment    start_align;
         PedAlignment    end_align;
         PedGeometry     max_geom;
 	SunDiskData*	disk_data = disk->disk_specific;
-	PedSector	block = disk->dev->sectors * disk->dev->heads;
+	PedSector	block = dev->sectors * dev->heads;
 
         if (!ped_alignment_init (&start_align, 0, block))
                 return NULL;
         if (!ped_alignment_init (&end_align, -1, block))
                 return NULL;
-	if (!ped_geometry_init (&max_geom, disk, 0, disk_data->length))
+	if (!ped_geometry_init (&max_geom, dev, 0, disk_data->length))
 		return NULL;
 
         return ped_constraint_new (&start_align, &end_align, &max_geom,
-                                   &max_geom, 1);
+                                   &max_geom, 1, dev->length);
 }
 
 static PedConstraint*
 _get_lax_constraint (PedDisk* disk)
 {
+	PedDevice*	dev = disk->dev;
         PedAlignment    start_align;
         PedGeometry     max_geom;
 	SunDiskData*	disk_data = disk->disk_specific;
-	PedSector	block = disk->dev->sectors * disk->dev->heads;
+	PedSector	block = dev->sectors * dev->heads;
 
         if (!ped_alignment_init (&start_align, 0, block))
                 return NULL;
-	if (!ped_geometry_init (&max_geom, disk, 0, disk_data->length))
+	if (!ped_geometry_init (&max_geom, dev, 0, disk_data->length))
 		return NULL;
 
         return ped_constraint_new (&start_align, ped_alignment_any, &max_geom,
-                                   &max_geom, 1);
-}
-
-/* argghh!  this should be somewhere else... but where?! */
-static int
-_try_constraint (PedPartition* part, const PedConstraint* external,
-                 PedConstraint* internal)
-{
-        PedConstraint*          intersection;
-        PedGeometry*            solution;
-
-        intersection = ped_constraint_intersect (external, internal);
-        ped_constraint_destroy (internal);
-        if (!intersection)
-                goto fail;
-
-        solution = ped_constraint_solve_nearest (intersection, &part->geom);
-        if (!solution)
-                goto fail_free_intersection;
-        ped_geometry_set (&part->geom, solution->start, solution->length);
-        ped_geometry_destroy (solution);
-        ped_constraint_destroy (intersection);
-        return 1;
-
-fail_free_intersection:
-        ped_constraint_destroy (intersection);
-fail:
-        return 0;
+                                   &max_geom, 1, dev->length);
 }
 
 /* _get_strict_constraint() will align the partition to the end of the cylinder.
@@ -741,17 +663,19 @@ sun_partition_align (PedPartition* part, const PedConstraint* constraint)
 {
         PED_ASSERT (part != NULL, return 0);
 
-        if (_try_constraint (part, constraint,
-                             _get_strict_constraint (part->geom.disk)))
-		return 1;
-        if (_try_constraint (part, constraint,
-                             _get_lax_constraint (part->geom.disk)))
-        	return 1;
+	if (_ped_partition_attempt_align (part, constraint,
+					  _get_strict_constraint (part->disk)))
+	       	return 1;
+	if (_ped_partition_attempt_align (part, constraint,
+					  _get_lax_constraint (part->disk)))
+	       	return 1;
 
+#ifndef DISCOVER_ONLY
 	ped_exception_throw (
 		PED_EXCEPTION_ERROR,
 		PED_EXCEPTION_CANCEL,
-		_("Unable to align partition."));
+		_("Unable to satisfy all constraints on the partition."));
+#endif
 	return 0;
 }
 
@@ -768,15 +692,16 @@ sun_partition_enumerate (PedPartition* part)
 		/* skip the Whole Disk partition for now */
 		if (i == WHOLE_DISK_PART + 1)
 			continue;
-		p = ped_disk_get_partition (part->geom.disk, i);
+		p = ped_disk_get_partition (part->disk, i);
 		if (!p) {
 			part->num = i;
 			return 1;
 		}
 	}
 
+#ifndef DISCOVER_ONLY
 	/* Ok, now allocate the Whole disk if it isn't already */
-	p = ped_disk_get_partition (part->geom.disk, WHOLE_DISK_PART + 1);
+	p = ped_disk_get_partition (part->disk, WHOLE_DISK_PART + 1);
 	if (!p) {
 		int i = ped_exception_throw (
 				PED_EXCEPTION_WARNING,
@@ -797,6 +722,7 @@ sun_partition_enumerate (PedPartition* part)
 	/* failed to allocate a number, this means we are full */
 	ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
 			     _("Sun disk label is full."));
+#endif
 	return 0;
 }
 
@@ -805,12 +731,14 @@ sun_alloc_metadata (PedDisk* disk)
 {
 	PedPartition*	new_part;
 	SunDiskData*	disk_data;
-	PedConstraint*	constraint_any = ped_constraint_any (disk);
+	PedConstraint*	constraint_any;
 
 	PED_ASSERT (disk != NULL, return 0);
 	PED_ASSERT (disk->disk_specific != NULL, return 0);
 	PED_ASSERT (disk->dev != NULL, return 0);
 
+	constraint_any = ped_constraint_any (disk->dev);
+
 	/* Sun disk label does not need to allocate a sector. The disk
 	   label is contained within the first 512 bytes, which should not
 	   be overwritten by any boot loader or superblock. It is safe for
@@ -825,11 +753,8 @@ sun_alloc_metadata (PedDisk* disk)
 	    disk->dev->length == disk_data->length)
 		goto error;
 
-	new_part = ped_partition_new (
-			disk,
-			PED_PARTITION_PRIMARY | PED_PARTITION_METADATA,
-			NULL, disk_data->length,
-			disk->dev->length - 1);
+	new_part = ped_partition_new (disk, PED_PARTITION_METADATA, NULL,
+			disk_data->length, disk->dev->length - 1);
 	if (!new_part)
 		goto error;
 
@@ -844,3 +769,58 @@ error:
 	ped_constraint_destroy (constraint_any);
 	return 0;
 }
+
+static PedDiskOps sun_disk_ops = {
+	probe:			sun_probe,
+#ifndef DISCOVER_ONLY
+	clobber:		sun_clobber,
+#else
+	clobber:		NULL,
+#endif
+	alloc:			sun_alloc,
+	duplicate:		sun_duplicate,
+	free:			sun_free,
+	read:			sun_read,
+#ifndef DISCOVER_ONLY
+	write:			sun_write,
+#else
+	write:			NULL,
+#endif
+
+	partition_new:		sun_partition_new,
+	partition_duplicate:	sun_partition_duplicate,
+	partition_destroy:	sun_partition_destroy,
+	partition_set_system:	sun_partition_set_system,
+	partition_set_flag:	sun_partition_set_flag,
+	partition_get_flag:	sun_partition_get_flag,
+	partition_is_flag_available:	sun_partition_is_flag_available,
+	partition_align:	sun_partition_align,
+	partition_enumerate:	sun_partition_enumerate,
+	alloc_metadata:		sun_alloc_metadata,
+	get_max_primary_partition_count:
+				sun_get_max_primary_partition_count,
+
+	partition_set_name:		NULL,
+	partition_get_name:		NULL,
+};
+
+static PedDiskType sun_disk_type = {
+	next:		NULL,
+	name:		"sun",
+	ops:		&sun_disk_ops,
+	features:	0
+};
+
+void
+ped_disk_sun_init ()
+{
+	PED_ASSERT (sizeof (SunRawLabel) == 512, return);
+	ped_register_disk_type (&sun_disk_type);
+}
+
+void
+ped_disk_sun_done ()
+{
+	ped_unregister_disk_type (&sun_disk_type);
+}
+
