@@ -1,10 +1,6 @@
 /*
     libparted - a library for manipulating disk partitions
-    Copyright (C) 1998-2000  Andrew Clausen, Lennert Buytenhek and Red Hat Inc.
-
-	Andrew Clausen			<clausen@gnu.org>
-	Lennert Buytenhek		<buytenh@gnu.org>
-	Matt Wilson, Red Hat Inc.	<msw@redhat.com>
+    Copyright (C) 1999, 2000 Free Software Foundation, Inc.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -24,11 +20,12 @@
 #include "config.h"
 
 #include <parted/parted.h>
+#include <parted/debug.h>
 
-#include <libintl.h>
 #define N_(String) String
 #if ENABLE_NLS
-#  define _(String) gettext (String)
+#  include <libintl.h>
+#  define _(String) dgettext (PACKAGE, String)
 #else
 #  define _(String) (String)
 #endif /* ENABLE_NLS */
@@ -88,14 +85,16 @@ static PedExceptionOption
 default_handler (PedException* ex)
 {
 	if (ex->type == PED_EXCEPTION_BUG)
-		printf (_("A bug has been detected in GNU parted.  "
+		fprintf (stderr,
+			_("A bug has been detected in GNU parted.  "
 			"Please email a bug report to "
 			"bug-parted@gnu.org containing the version (%s) "
 			"and the following message:"),
 			VERSION);
 	else
-		printf ("%s: ", ped_exception_get_type_string (ex->type));
-	printf ("%s\n", ex->message);
+		fprintf (stderr, "%s: ",
+			 ped_exception_get_type_string (ex->type));
+	fprintf (stderr, "%s\n", ex->message);
 
 	switch (ex->options) {
 		case PED_EXCEPTION_OK:
@@ -111,7 +110,10 @@ default_handler (PedException* ex)
 void
 ped_exception_set_handler (PedExceptionHandler* handler)
 {
-	ex_handler = handler;
+	if (handler)
+		ex_handler = handler;
+	else
+		ex_handler = default_handler;
 }
 
 void
@@ -148,6 +150,9 @@ ped_exception_throw (PedExceptionType ex_type,
 {
 	va_list		arg_list;
 
+	if (ex)
+		ped_exception_catch ();
+
 	ex = (PedException*) malloc (sizeof (PedException));
 	if (!ex)
 		goto no_memory;
@@ -166,10 +171,10 @@ ped_exception_throw (PedExceptionType ex_type,
 	return do_throw ();
 
 no_memory:
-	printf ("Out of memory in exception handler!\n");
+	fprintf (stderr, "Out of memory in exception handler!\n");
 
 	va_start (arg_list, message);
-	vprintf (message, arg_list);
+	vfprintf (stderr, message, arg_list);
 	va_end (arg_list);
 
 	return PED_EXCEPTION_UNHANDLED;
@@ -190,5 +195,6 @@ ped_exception_fetch_all ()
 void
 ped_exception_leave_all ()
 {
+	PED_ASSERT (ex_fetch_count > 0, return);
 	ex_fetch_count--;
 }
