@@ -17,10 +17,10 @@
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
-static const char _parted_io_c[] = "$Id: parted_io.c,v 1.4 1999/10/02 20:56:47 aclausen Exp $";
-
 #include "config.h"
 
+#ifndef DISCOVER_ONLY
+
 #include <parted/parted.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -133,3 +133,4 @@ void ext2_destroy_dev_handle(struct ext2_dev_handle *handle)
 	ped_free(handle->cookie);
 	ped_free(handle);
 }
+#endif /* !DISCOVER_ONLY */
