@@ -197,7 +197,7 @@ flag_traverse_fat (PedFileSystem* fs, const char* chain_name, FatCluster start,
 		if (ped_exception_throw (
 			PED_EXCEPTION_ERROR,
 			PED_EXCEPTION_IGNORE_CANCEL,
-			_("%s is %dk, but it has is %d clusters (%dk)."),
+			_("%s is %dk, but it has %d clusters (%dk)."),
 			chain_name,
 			(int) size / 2,
 			(int) chain_length,
