@@ -20,51 +20,9 @@
 #include <string.h>
 #include <uuid/uuid.h>
 
-#include <parted/disk_dos.h>
-#include <parted/disk_loop.h>
-#include <parted/disk_pc98.h>
-#include <parted/disk_mac.h>
-#include <parted/disk_gpt.h>
-
 #include "fat.h"
 #include "calc.h"
 
-static PedFileSystemOps fat_ops = {
-	probe:		fat_probe,
-	clobber:	fat_clobber,
-	open:		fat_open,
-	create:		fat_create,
-	close:		fat_close,
-	check:		fat_check,
-	resize:		fat_resize,
-	copy:		fat_copy,
-	get_resize_constraint:	fat_get_resize_constraint,
-	set_system:	fat_set_system
-};
-
-static PedFileSystemType fat_type = {
-	next:	NULL,
-	ops:	&fat_ops,
-	name:	"FAT"
-};
-
-void
-ped_file_system_fat_init ()
-{
-	if (sizeof (FatBootSector) != 512)
-		ped_exception_throw (PED_EXCEPTION_BUG, PED_EXCEPTION_CANCEL,
-			_("GNU parted was miscompiled: the FAT boot sector "
-			"should be 512 bytes.  FAT support will be disabled."));
-	else
-		ped_file_system_type_register (&fat_type);
-}
-
-void
-ped_file_system_fat_done ()
-{
-	ped_file_system_type_unregister (&fat_type);
-}
-
 PedFileSystem*
 fat_alloc (const PedGeometry* geom)
 {
@@ -81,7 +39,8 @@ fat_alloc (const PedGeometry* geom)
 	fs->geom = ped_geometry_duplicate (geom);
 	if (!fs->geom)
 		goto error_free_type_specific;
-	
+
+	fs->checked = 0;
 	return fs;
 
 error_free_type_specific:
@@ -150,14 +109,69 @@ fat_set_frag_sectors (PedFileSystem* fs, PedSector frag_sectors)
 	return 1;
 }
 
-int
-fat_probe (const PedGeometry* geom)
+PedGeometry*
+fat_probe (PedGeometry* geom, FatType* fat_type)
 {
-	FatBootSector		boot_sector;
+	PedFileSystem*		fs;
+	FatSpecific*		fs_info;
+	PedGeometry*		result;
 
-	return fat_boot_sector_read (&boot_sector, geom);
+	fs = fat_alloc (geom);
+	if (!fs)
+		goto error;
+	fs_info = (FatSpecific*) fs->type_specific;
+
+	if (!fat_boot_sector_read (&fs_info->boot_sector, geom))
+		goto error_free_fs;
+	if (!fat_boot_sector_analyse (&fs_info->boot_sector, fs)) {
+		*fat_type = fat_boot_sector_probe_type (&fs_info->boot_sector,
+							geom);
+		fat_free (fs);
+		return ped_geometry_duplicate (geom);
+	}
+
+	*fat_type = fs_info->fat_type;
+	result = ped_geometry_new (geom->dev, geom->start,
+				   fs_info->sector_count);
+
+	fat_free (fs);
+	return result;
+
+error_free_fs:
+	fat_free (fs);
+error:
+	return NULL;
 }
 
+PedGeometry*
+fat_probe_fat16 (PedGeometry* geom)
+{
+	FatType		fat_type;
+	PedGeometry*	probed_geom = fat_probe (geom, &fat_type);
+
+	if (probed_geom) {
+		if (fat_type == FAT_TYPE_FAT16)
+			return probed_geom;
+		ped_geometry_destroy (probed_geom);
+	}
+	return NULL;
+}
+
+PedGeometry*
+fat_probe_fat32 (PedGeometry* geom)
+{
+	FatType		fat_type;
+	PedGeometry*	probed_geom = fat_probe (geom, &fat_type);
+
+	if (probed_geom) {
+		if (fat_type == FAT_TYPE_FAT32)
+			return probed_geom;
+		ped_geometry_destroy (probed_geom);
+	}
+	return NULL;
+}
+
+#ifndef DISCOVER_ONLY
 int
 fat_clobber (PedGeometry* geom)
 {
@@ -168,6 +182,11 @@ fat_clobber (PedGeometry* geom)
 
 	boot_sector.system_id[0] = 0;
 	boot_sector.boot_sign = 0;
+	if (boot_sector.u.fat16.fat_name[0] == 'F')
+		boot_sector.u.fat16.fat_name[0] = 0;
+	if (boot_sector.u.fat32.fat_name[0] == 'F')
+		boot_sector.u.fat32.fat_name[0] = 0;
+
         return ped_geometry_write (geom, &boot_sector, 0, 1);
 }
 
@@ -177,39 +196,18 @@ _init_fats (PedFileSystem* fs)
 	FatSpecific*	fs_info = FAT_SPECIFIC (fs);
 	int		i;
 	FatCluster	table_size;
-	FatTable*	fat_copy;
 
 	table_size = fs_info->fat_sectors * 512
 		     / fat_table_entry_size (fs_info->fat_type);
 	fs_info->fat = fat_table_new (fs_info->fat_type, table_size);
-	if (!fs_info->fat) goto error;
-
-	fat_copy = fat_table_new (fs_info->fat_type, table_size);
-	if (!fat_copy) goto error_free_fat;
+	if (!fs_info->fat)
+		goto error;
 
 	if (!fat_table_read (fs_info->fat, fs, 0))
-		goto error_free_copy;
+		goto error_free_fat;
 
-	for (i = 1; i < fs_info->fat_table_count; i++) {
-		if (!fat_table_read (fat_copy, fs, i))
-			goto error_free_copy;
-		if (!fat_table_compare (fs_info->fat, fat_copy)) {
-			if (ped_exception_throw (PED_EXCEPTION_ERROR,
-				PED_EXCEPTION_IGNORE_CANCEL,
-				_("The FATs don't match.  If you don't know "
-				"what this means, then select cancel, run "
-				"scandisk on the file system, and then come "
-				"back."))
-			    != PED_EXCEPTION_IGNORE)
-				goto error_free_copy;
-		}
-	}
-
-	fat_table_destroy (fat_copy);
 	return 1;
 
-error_free_copy:
-	fat_table_destroy (fat_copy);
 error_free_fat:
 	fat_table_destroy (fs_info->fat);
 error:
@@ -217,7 +215,7 @@ error:
 }
 
 PedFileSystem*
-fat_open (const PedGeometry* geom)
+fat_open (PedGeometry* geom)
 {
 	PedFileSystem*		fs;
 	FatSpecific*		fs_info;
@@ -227,12 +225,13 @@ fat_open (const PedGeometry* geom)
 		goto error;
 	fs_info = (FatSpecific*) fs->type_specific;
 
-	fs->type = &fat_type;
-
 	if (!fat_boot_sector_read (&fs_info->boot_sector, geom))
 		goto error_free_fs;
 	if (!fat_boot_sector_analyse (&fs_info->boot_sector, fs))
 		goto error_free_fs;
+	fs->type = (fs_info->fat_type == FAT_TYPE_FAT16)
+				? &fat16_type
+				: &fat32_type;
 	if (fs_info->fat_type == FAT_TYPE_FAT32) {
 		if (!fat_info_sector_read (&fs_info->info_sector, fs))
 			goto error_free_fs;
@@ -270,17 +269,17 @@ fat_root_dir_clear (PedFileSystem* fs)
 /* hack: use the ext2 uuid library to generate a reasonably random (hopefully
  * with /dev/random) number.  Unfortunately, we can only use 4 bytes of it
  */
-static __u32
+static uint32_t
 _gen_new_serial_number ()
 {
 	uuid_t		uuid;
 
 	uuid_generate (uuid);
-	return * (__u32*) &uuid [0];
+	return * (uint32_t*) &uuid [0];
 }
 
 PedFileSystem*
-fat_create (PedGeometry* geom)
+fat_create (PedGeometry* geom, FatType fat_type, PedTimer* timer)
 {
 	PedFileSystem*		fs;
 	FatSpecific*		fs_info;
@@ -291,8 +290,6 @@ fat_create (PedGeometry* geom)
 		goto error;
 	fs_info = (FatSpecific*) fs->type_specific;
 
-	fs->type = &fat_type;
-
 	fs_info->logical_sector_size = 1;
 	fs_info->sector_count = fs->geom->length;
 	fs_info->fat_table_count = 2;
@@ -301,82 +298,20 @@ fat_create (PedGeometry* geom)
 					  / (512 / sizeof (FatDirEntry));
 	fs_info->root_dir_entry_count = FAT_ROOT_DIR_ENTRY_COUNT;
 
-	fs_info->fat_type = FAT_TYPE_FAT16;
-	if (!fat_calc_sizes (fs->geom, 0,
-			FAT_TYPE_FAT16,
+	fs_info->fat_type = fat_type;
+	if (!fat_calc_sizes (fs->geom->length, 0,
+			fs_info->fat_type,
 			fs_info->root_dir_sector_count,
 			&fs_info->cluster_sectors,
 			&fs_info->cluster_count,
 			&fs_info->fat_sectors)) {
-		if (!fat_calc_sizes (fs->geom, 0,
-				     FAT_TYPE_FAT32,
-				     0,
-				     &fs_info->cluster_sectors,
-				     &fs_info->cluster_count,
-				     &fs_info->fat_sectors)) {
-			ped_exception_throw (PED_EXCEPTION_ERROR,
-				PED_EXCEPTION_CANCEL,
-				_("Partition too small for a FAT file system"));
-			goto error_free_fs;
-		}
-		if (ped_exception_throw (PED_EXCEPTION_WARNING,
-			PED_EXCEPTION_OK_CANCEL,
-			_("The filesystem is going to be too big for FAT16, so "
-			  "FAT32 will be used.  This is not compatible with "
-			  "MS-DOS, early versions of MS-Windows 95 and "
-			  "Windows NT.  If you use these operating systems, "
-			  "then select cancel, and create a smaller "
-			  "partition.  If you only use Linux, BSD, MS "
-			  "Windows 98 and/or MS Windows 95 B, then select OK."))
-				== PED_EXCEPTION_CANCEL)
-			goto error_free_fs;
-		if (!fat_calc_sizes (fs->geom, 0,
-				     FAT_TYPE_FAT32,
-				     0,
-				     &fs_info->cluster_sectors,
-				     &fs_info->cluster_count,
-				     &fs_info->fat_sectors)) {
-			ped_exception_throw (PED_EXCEPTION_BUG,
-				PED_EXCEPTION_CANCEL,
-				_("Weird: fat_calc_sizes() failed "
-				"for FAT32!"));
-			goto error_free_fs;
-		}
-		fs_info->fat_type = FAT_TYPE_FAT32;
-	}
-
-	if (fs_info->cluster_sectors * 512 > 8192) {
-		switch (ped_exception_throw (PED_EXCEPTION_WARNING,
-			PED_EXCEPTION_YES_NO_CANCEL,
-			_("Would you like to use FAT32 for this filesystem?  "
-			"It is much more efficient with your disk space, but "
-			"is not compatible with early versions of Windows 95 "
-			"and Windows NT.  Only select yes if you only "
-			"use Linux, BSD, MS Windows 98 and/or MS Windows 95 "
-			"B."))) {
-		case PED_EXCEPTION_CANCEL:
-			goto error_free_fs;
-
-		case PED_EXCEPTION_YES:
-			if (!fat_calc_sizes (fs->geom, 0,
-					     FAT_TYPE_FAT32,
-					     0,
-					     &fs_info->cluster_sectors,
-					     &fs_info->cluster_count,
-					     &fs_info->fat_sectors)) {
-				ped_exception_throw (PED_EXCEPTION_BUG,
-					PED_EXCEPTION_CANCEL,
-					_("Weird: fat_calc_sizes() failed "
-					"for FAT32!"));
-				goto error_free_fs;
-			}
-			fs_info->fat_type = FAT_TYPE_FAT32;
-			break;
-
-		case PED_EXCEPTION_NO:
-		default:
-			break;
-		}
+		ped_exception_throw (PED_EXCEPTION_ERROR,
+			PED_EXCEPTION_CANCEL,
+			_("Partition too big/small for a %s file system"),
+			(fat_type == FAT_TYPE_FAT16)
+		       		? fat16_type.name
+				: fat32_type.name);
+		goto error_free_fs;
 	}
 
 	fs_info->cluster_size = fs_info->cluster_sectors * 512;
@@ -387,6 +322,8 @@ fat_create (PedGeometry* geom)
 
 	if (fs_info->fat_type == FAT_TYPE_FAT16) {
 		/* FAT16 */
+		fs->type = &fat16_type;
+
 		if (fs_info->cluster_count
 			> fat_max_cluster_count (fs_info->fat_type)) {
 			fs_info->cluster_count
@@ -405,6 +342,8 @@ fat_create (PedGeometry* geom)
 			  + fs_info->root_dir_sector_count;
 	} else {
 		/* FAT32 */
+		fs->type = &fat32_type;
+
 		fs_info->info_sector_offset = 1;
 		fs_info->boot_sector_backup_offset = 6;
 
@@ -422,6 +361,7 @@ fat_create (PedGeometry* geom)
 	fs_info->fat = fat_table_new (fs_info->fat_type, table_size);
 	if (!fs_info->fat)
 		goto error_free_fs;
+	fat_table_set_cluster_count (fs_info->fat, fs_info->cluster_count);
 	if (!fat_alloc_buffers (fs)) 
 		goto error_free_fat_table;
 
@@ -470,6 +410,18 @@ error:
 	return NULL;
 }
 
+PedFileSystem*
+fat_create_fat16 (PedGeometry* geom, PedTimer* timer)
+{
+	return fat_create (geom, FAT_TYPE_FAT16, timer);
+}
+
+PedFileSystem*
+fat_create_fat32 (PedGeometry* geom, PedTimer* timer)
+{
+	return fat_create (geom, FAT_TYPE_FAT32, timer);
+}
+
 int
 fat_close (PedFileSystem* fs)
 {
@@ -482,27 +434,65 @@ fat_close (PedFileSystem* fs)
 }
 
 /* Hack: just resize the filesystem outside of it's boundaries! */
-int
-fat_copy (const PedFileSystem* fs, PedGeometry* geom)
+PedFileSystem*
+fat_copy (const PedFileSystem* fs, PedGeometry* geom, PedTimer* timer)
 {
 	PedFileSystem*		new_fs;
 
-	new_fs = fat_open (fs->geom);
+	new_fs = ped_file_system_open (fs->geom);
 	if (!new_fs)
 		goto error;
-	if (!fat_resize (new_fs, geom))
+	if (!ped_file_system_resize (new_fs, geom, timer))
 		goto error_close_new_fs;
-	fat_close (new_fs);
-	return 1;
+	return new_fs;
 
 error_close_new_fs:
-	fat_close (new_fs);
+	ped_file_system_close (new_fs);
+error:
+	return 0;
+}
+
+static int
+_compare_fats (PedFileSystem* fs)
+{
+	FatSpecific*	fs_info = FAT_SPECIFIC (fs);
+	FatTable*	table_copy;
+	FatCluster	table_size;
+	int		i;
+
+	table_size = fs_info->fat_sectors * 512
+		     / fat_table_entry_size (fs_info->fat_type);
+
+	table_copy = fat_table_new (fs_info->fat_type, table_size);
+	if (!table_copy)
+		goto error;
+
+	for (i = 1; i < fs_info->fat_table_count; i++) {
+		if (!fat_table_read (table_copy, fs, i))
+			goto error_free_table_copy;
+		if (!fat_table_compare (fs_info->fat, table_copy)) {
+			if (ped_exception_throw (PED_EXCEPTION_ERROR,
+				PED_EXCEPTION_IGNORE_CANCEL,
+				_("The FATs don't match.  If you don't know "
+				"what this means, then select cancel, run "
+				"scandisk on the file system, and then come "
+				"back."))
+			    != PED_EXCEPTION_IGNORE)
+				goto error_free_table_copy;
+		}
+	}
+
+	fat_table_destroy (table_copy);
+	return 1;
+
+error_free_table_copy:
+	fat_table_destroy (table_copy);
 error:
 	return 0;
 }
 
 int
-fat_check (PedFileSystem* fs)
+fat_check (PedFileSystem* fs, PedTimer* timer)
 {
 	FatSpecific*	fs_info = FAT_SPECIFIC (fs);
 	PedSector	cluster_sectors;
@@ -514,7 +504,7 @@ fat_check (PedFileSystem* fs)
 	align_sectors = fs_info->fat_offset
 			- fat_min_reserved_sector_count (fs_info->fat_type);
 
-	if (!fat_calc_sizes (fs->geom,
+	if (!fat_calc_sizes (fs->geom->length,
 			     align_sectors,
 			     fs_info->fat_type,
 			     fs_info->root_dir_sector_count,
@@ -526,7 +516,7 @@ fat_check (PedFileSystem* fs)
 			_("There are no possible configurations for this FAT "
 			  "type."))
 				!= PED_EXCEPTION_IGNORE)
-			return 0;
+			goto error;
 	}
 
 	if (fs_info->fat_type == FAT_TYPE_FAT16) {
@@ -547,27 +537,9 @@ fat_check (PedFileSystem* fs)
 				(int) fs_info->fat_sectors,
 					(int) fat_sectors)
 					!= PED_EXCEPTION_IGNORE)
-				return 0;
+				goto error;
 		}
-	} else {
-#if 0
-		if (cluster_count != fs_info->cluster_count
-		    || fat_sectors != fs_info->fat_sectors) {
-			if (ped_exception_throw (PED_EXCEPTION_WARNING,
-				PED_EXCEPTION_IGNORE_CANCEL,
-				_("File system doesn't have expected sizes for "
-				"Windows to like it.  "
-				"Number of clusters is %d (%d expected); "
-				"size of FATs is %d sectors (%d expected)."),
-				(int) fs_info->cluster_count,
-					(int) cluster_count,
-				(int) fs_info->fat_sectors,
-					(int) fat_sectors)
-					!= PED_EXCEPTION_IGNORE)
-				return 0;
-		}
-#endif
-	}
+	} 
 
 	if (fs_info->fat_type == FAT_TYPE_FAT32) {
 		info_free_clusters
@@ -580,11 +552,18 @@ fat_check (PedFileSystem* fs)
 				info_free_clusters,
 				fs_info->fat->free_cluster_count)
 					!= PED_EXCEPTION_IGNORE)
-				return 0;
+				goto error;
 		}
 	}
 
+	if (!_compare_fats (fs))
+		goto error;
+
+	fs->checked = 1;
 	return 1;	/* existence of fs implies consistency ;-) */
+
+error:
+	return 0;
 }
 
 /* Calculates how much space there will be in clusters in:
@@ -605,7 +584,8 @@ _calc_resize_data_size (
 }
 
 static int
-_test_size (const PedFileSystem* fs, PedSector length, PedSector min_data_size)
+_test_resize_size (const PedFileSystem* fs,
+		   PedSector length, PedSector min_data_size)
 {
 	FatSpecific*	fs_info = FAT_SPECIFIC (fs);
 	PedGeometry	geom;
@@ -613,7 +593,7 @@ _test_size (const PedFileSystem* fs, PedSector length, PedSector min_data_size)
 	FatCluster	_cluster_count;
 	PedSector	_fat_size;
 
-	ped_geometry_init (&geom, fs->geom->disk, fs->geom->start, length);
+	ped_geometry_init (&geom, fs->geom->dev, fs->geom->start, length);
 
 	if (fat_calc_resize_sizes (
 				&geom,
@@ -650,7 +630,7 @@ _test_size (const PedFileSystem* fs, PedSector length, PedSector min_data_size)
  * (see calc_sizes() for why!)
  */
 static PedSector
-_get_min_size (const PedFileSystem* fs, PedSector min_data_size)
+_get_min_resize_size (const PedFileSystem* fs, PedSector min_data_size)
 {
 	PedSector	min_length = 0;
 	PedSector	max_length = fs->geom->length;
@@ -658,7 +638,7 @@ _get_min_size (const PedFileSystem* fs, PedSector min_data_size)
 
 	while (min_length < max_length - 1) {
 		length = (min_length + max_length) / 2;
-		if (_test_size (fs, length, min_data_size))
+		if (_test_resize_size (fs, length, min_data_size))
 			max_length = length;
 		else
 			min_length = length;
@@ -671,16 +651,15 @@ _get_min_size (const PedFileSystem* fs, PedSector min_data_size)
 }
 
 PedConstraint*
-fat_get_resize_constraint (const PedFileSystem* fs)
+fat_get_copy_constraint (const PedFileSystem* fs, const PedDevice* dev)
 {
 	FatSpecific*	fs_info = FAT_SPECIFIC (fs);
-	PedDisk*	disk = fs->geom->disk;
-	PedGeometry	full_disk;
+	PedGeometry	full_dev;
 	PedSector	min_cluster_count;
 	FatCluster	used_clusters;
 	PedSector	min_data_size;
 
-	if (!ped_geometry_init (&full_disk, disk, 0, disk->dev->length - 1))
+	if (!ped_geometry_init (&full_dev, dev, 0, dev->length - 1))
 		return NULL;
 
 	used_clusters = fs_info->fat->cluster_count
@@ -689,118 +668,218 @@ fat_get_resize_constraint (const PedFileSystem* fs)
 	min_data_size = min_cluster_count * fs_info->cluster_sectors;
 
 	return ped_constraint_new (ped_alignment_any, ped_alignment_any,
-				   &full_disk,
-				   &full_disk,
-				   _get_min_size (fs, min_data_size));
+				   &full_dev, &full_dev,
+				   _get_min_resize_size (fs, min_data_size),
+				   dev->length);
 }
 
-int
-fat_set_system (const PedFileSystem* fs, PedPartition* part,
-		const PedDiskType* disk_type)
+PedConstraint*
+fat_get_resize_constraint (const PedFileSystem* fs)
 {
-	PedDevice*		dev;
+	return fat_get_copy_constraint (fs, fs->geom->dev);
+}
 
-	dev = part->geom.disk->dev;
+/* FIXME: fat_calc_sizes() needs to say "too big" or "too small", or
+ * something.  This is a really difficult (maths) problem to do
+ * nicely...
+ * 	So, this algorithm works if dev->length / 2 is a valid fat_type
+ * size.  (Which is how I got the magic numbers below)
+ */
+#if 0
+/* returns: -1 too small, 0 ok, 1 too big */
+static int
+_test_create_size (PedSector length, FatType fat_type,
+		   PedSector cluster_sectors, PedSector cluster_count)
+{
+	PedSector	rootdir_sectors;
+	PedSector	_cluster_sectors;
+	FatCluster	_cluster_count;
+	PedSector	_fat_size;
 
-	if (strcasecmp (disk_type->name, LOOP_NAME) == 0)
+	rootdir_sectors = (fat_type == FAT_TYPE_FAT16) ? 16 : 0;
+
+	if (!fat_calc_sizes (length, 0, fat_type, rootdir_sectors,
+			     &_cluster_sectors, &_cluster_count, &_fat_size))
+		return -1; // XXX: doesn't work... can't see a better way!
+
+	if (_cluster_sectors < cluster_sectors)
+		return -1;
+	if (_cluster_sectors > cluster_sectors)
 		return 1;
 
-	if (strcasecmp (disk_type->name, DOS_NAME) == 0) {
-		DosPartitionData*	dos_data = part->disk_specific;
-
-		PED_ASSERT (dos_data != NULL, return 0);
-
-		if (part->geom.end >= 1024 * dev->heads * dev->sectors)
-			dos_data->lba = 1;
-
-		if (dos_data->raid) {
-			dos_data->system = 0xfd;
-			return 1;
-		}
-		if (dos_data->lvm) {
-			dos_data->system = 0x8e;
-			return 1;
-		}
-
-		if (!fs) {
-			if (dos_data->hidden)
-				dos_data->system = dos_data->lba ? 0x1e : 0x16;
-			else
-				dos_data->system = dos_data->lba ? 0x0e : 0x06;
-			return 1;
-		}
-
-		if (FAT_SPECIFIC (fs)->fat_type == FAT_TYPE_FAT32) {
-			if (dos_data->hidden)
-				dos_data->system = dos_data->lba ? 0x1c : 0x1b;
-			else
-				dos_data->system = dos_data->lba ? 0x0c : 0x0b;
-		} else {
-			if (dos_data->hidden)
-				dos_data->system = dos_data->lba ? 0x1e : 0x16;
-			else
-				dos_data->system = dos_data->lba ? 0x0e : 0x06;
-		}
+	if (_cluster_count < cluster_count)
+		return -1;
+	if (_cluster_count > cluster_count)
 		return 1;
-	}
-
-	if (strcasecmp (disk_type->name, PC98_NAME) == 0) {
-		PC98PartitionData*	pc98_data = part->disk_specific;
-
-		PED_ASSERT (pc98_data != NULL, return 0);
-
-		if (!fs) {
-			pc98_data->system = 0x2021;
-			return 1;
-		} 
-
-		if ((pc98_data->system & 0x7000) != 0x2000
-		    || !pc98_data->boot
-		    || (FAT_SPECIFIC(fs)->fat_type == FAT_TYPE_FAT32))
-			pc98_data->system = (pc98_data->system & 0xFF) | 0x2000;
-
-		pc98_data->system &= 0xFF00;
-		switch (FAT_SPECIFIC(fs)->fat_type) {
-		case FAT_TYPE_FAT32:
-			pc98_data->system |= 0x61;
-			break;
-		case FAT_TYPE_FAT12:
-			pc98_data->system |= 0x01;
-			break;
-		case FAT_TYPE_FAT16:
-			PED_ASSERT(part != NULL, return 0);
-			if (part->geom.length * PED_SECTOR_SIZE >=
-			    32 * 1024 * 1024)
-				pc98_data->system |= 0x21;
-			else
-				pc98_data->system |= 0x11;
-			break;
-		}
-		return 1;
-	}
-
-	if (strcasecmp (disk_type->name, MAC_NAME) == 0) {
-		MacPartitionData*	mac_data = part->disk_specific;
-
-		PED_ASSERT (mac_data != NULL, return 0);
-
-		if (mac_data->is_boot)
-			strcpy (mac_data->system_name, "Apple_Bootstrap");
-		else
-			strcpy (mac_data->system_name, "FAT");
-		mac_data->status = 0x33;
-		return 1;
-	}
-
-	if (strcmp (disk_type->name, GPT_NAME) == 0) {
-		GPTPartitionData *gpt_part_data = part->disk_specific;
-		PED_ASSERT(gpt_part_data      != NULL, return 0);	
-		PED_ASSERT(gpt_part_data->pte != NULL, return 0);
-		gpt_part_data->pte->PartitionTypeGuid =
-			PARTITION_BASIC_DATA_GUID;
-		return 1;
-	}
 
 	return 0;
 }
 
+static PedSector
+_get_create_size (PedSector upper_bound, FatType fat_type,
+		  PedSector cluster_sectors, FatCluster cluster_count)
+{
+	PedSector	min_length = 0;
+	PedSector	max_length = upper_bound;
+	PedSector	length;
+
+	while (1) {
+		length = (min_length + max_length) / 2;
+		switch (_test_create_size (length, fat_type, cluster_sectors,
+					   cluster_count)) {
+			case -1: min_length = length; break;
+			case 0: return length;
+			case 1: max_length = length; break;
+		}
+		/* hack... won't always be able to get max cluster count
+		 * with max cluster size, etc. */
+		if (max_length - min_length == 1)
+			return min_length;
+	}
+
+	return 0;	/* shut gcc up */
+}
+#endif
+
+PedConstraint*
+fat_get_create_constraint_fat16 (const PedDevice* dev)
+{
+	PedGeometry	full_dev;
+	PedSector	min_size;
+	PedSector	max_size;
+
+	if (!ped_geometry_init (&full_dev, dev, 0, dev->length - 1))
+		return NULL;
+
+#if 0
+	min_size = _get_create_size (dev->length, FAT_TYPE_FAT16,
+		       		     fat_min_cluster_size (FAT_TYPE_FAT16),
+				     fat_min_cluster_count (FAT_TYPE_FAT16));
+	max_size = _get_create_size (dev->length, FAT_TYPE_FAT16,
+				     fat_max_cluster_size (FAT_TYPE_FAT16),
+				     fat_max_cluster_count (FAT_TYPE_FAT16));
+	if (!min_size)
+		return NULL;
+#else
+	min_size = 65794;
+	max_size = 2097153;
+#endif
+
+	return ped_constraint_new (
+			ped_alignment_any, ped_alignment_any,
+			&full_dev, &full_dev,
+			min_size, max_size);
+}
+
+PedConstraint*
+fat_get_create_constraint_fat32 (const PedDevice* dev)
+{
+	PedGeometry	full_dev;
+	PedSector	min_size;
+
+	if (!ped_geometry_init (&full_dev, dev, 0, dev->length - 1))
+		return NULL;
+
+#if 0
+	min_size = _get_create_size (dev->length, FAT_TYPE_FAT32,
+		       		     fat_min_cluster_size (FAT_TYPE_FAT32),
+				     fat_min_cluster_count (FAT_TYPE_FAT32));
+	if (!min_size)
+		return NULL;
+#else
+	min_size = 525224;
+#endif
+
+	return ped_constraint_new (
+			ped_alignment_any, ped_alignment_any,
+			&full_dev, &full_dev,
+			min_size, dev->length);
+}
+#endif /* !DISCOVER_ONLY */
+
+static PedFileSystemOps fat16_ops = {
+	probe:		fat_probe_fat16,
+#ifndef DISCOVER_ONLY
+	clobber:	fat_clobber,
+	open:		fat_open,
+	create:		fat_create_fat16,
+	close:		fat_close,
+	check:		fat_check,
+	resize:		fat_resize,
+	copy:		fat_copy,
+	get_create_constraint:	fat_get_create_constraint_fat16,
+	get_resize_constraint:	fat_get_resize_constraint,
+	get_copy_constraint:	fat_get_copy_constraint,
+#else /* !DISCOVER_ONLY */
+	clobber:	NULL,
+	open:		NULL,
+	create:		NULL,
+	close:		NULL,
+	check:		NULL,
+	resize:		NULL,
+	copy:		NULL,
+	get_create_constraint:	NULL,
+	get_resize_constraint:	NULL,
+	get_copy_constraint:	NULL,
+#endif /* !DISCOVER_ONLY */
+};
+
+static PedFileSystemOps fat32_ops = {
+	probe:		fat_probe_fat32,
+#ifndef DISCOVER_ONLY
+	clobber:	fat_clobber,
+	open:		fat_open,
+	create:		fat_create_fat32,
+	close:		fat_close,
+	check:		fat_check,
+	resize:		fat_resize,
+	copy:		fat_copy,
+	get_create_constraint:	fat_get_create_constraint_fat32,
+	get_resize_constraint:	fat_get_resize_constraint,
+	get_copy_constraint:	fat_get_copy_constraint,
+#else /* !DISCOVER_ONLY */
+	clobber:	NULL,
+	open:		NULL,
+	create:		NULL,
+	close:		NULL,
+	check:		NULL,
+	resize:		NULL,
+	copy:		NULL,
+	get_create_constraint:	NULL,
+	get_resize_constraint:	NULL,
+	get_copy_constraint:	NULL,
+#endif /* !DISCOVER_ONLY */
+};
+
+PedFileSystemType fat16_type = {
+	next:	NULL,
+	ops:	&fat16_ops,
+	name:	"fat16"
+};
+
+PedFileSystemType fat32_type = {
+	next:	NULL,
+	ops:	&fat32_ops,
+	name:	"fat32"
+};
+
+void
+ped_file_system_fat_init ()
+{
+	if (sizeof (FatBootSector) != 512) {
+		ped_exception_throw (PED_EXCEPTION_BUG, PED_EXCEPTION_CANCEL,
+			_("GNU parted was miscompiled: the FAT boot sector "
+			"should be 512 bytes.  FAT support will be disabled."));
+	} else {
+		ped_file_system_type_register (&fat16_type);
+		ped_file_system_type_register (&fat32_type);
+	}
+}
+
+void
+ped_file_system_fat_done ()
+{
+	ped_file_system_type_unregister (&fat16_type);
+	ped_file_system_type_unregister (&fat32_type);
+}
+
