@@ -33,6 +33,8 @@
 #include <stdarg.h>
 #include <string.h>
 
+#ifndef DISCOVER_ONLY
+
 /* Recursively builds (i.e. makes consistent) the duplicated directory tree
  * (leaving the original directory tree in tact)
  */
@@ -224,9 +226,9 @@ alloc_root_dir (FatOpContext* ctx)
 
 	PED_ASSERT (new_fs_info->fat_type == FAT_TYPE_FAT32, return 0);
 
-	cluster_count = div_round_up (
-			PED_MAX (16, old_fs_info->root_dir_sector_count),
-			new_fs_info->cluster_sectors);
+	cluster_count = ped_div_round_up (
+			   PED_MAX (16, old_fs_info->root_dir_sector_count),
+			   new_fs_info->cluster_sectors);
 
 	for (i = 0; i < cluster_count; i++) {
 		cluster = fat_table_alloc_check_cluster (new_fs_info->fat,
@@ -241,6 +243,43 @@ alloc_root_dir (FatOpContext* ctx)
 	return 1;
 }
 
+/* when converting FAT32 -> FAT16
+ * fat_duplicate clusters() duplicated the root directory unnecessarily.
+ * Let's free it.
+ *
+ * This must be called AFTER fat_construct_new_fat().  (otherwise, our
+ * changes just get overwritten)
+ */
+static int
+free_root_dir (FatOpContext* ctx)
+{
+	FatSpecific*		old_fs_info = FAT_SPECIFIC (ctx->old_fs);
+	FatSpecific*		new_fs_info = FAT_SPECIFIC (ctx->new_fs);
+	FatCluster		old_cluster;
+	FatFragment		i;
+
+	PED_ASSERT (old_fs_info->fat_type == FAT_TYPE_FAT32, return 0);
+	PED_ASSERT (new_fs_info->fat_type == FAT_TYPE_FAT16, return 0);
+
+	for (old_cluster = old_fs_info->root_cluster;
+	     !fat_table_is_eof (old_fs_info->fat, old_cluster);
+	     old_cluster = fat_table_get (old_fs_info->fat, old_cluster)) {
+		FatFragment old_frag;
+		old_frag = fat_cluster_to_frag (ctx->old_fs, old_cluster);
+		for (i = 0; i < new_fs_info->cluster_frags; i++) {
+			FatFragment new_frag;
+			FatCluster new_clst;
+			new_frag = fat_op_context_map_fragment (ctx,
+								old_frag + i);
+			new_clst = fat_frag_to_cluster (ctx->old_fs, new_frag);
+			if (!fat_table_set_avail (new_fs_info->fat, new_clst))
+				return 0;
+		}
+	}
+
+	return 1;
+}
+
 static int
 fat_clear_root_dir (PedFileSystem* fs)
 {
@@ -397,7 +436,7 @@ fat_construct_new_fat (FatOpContext* ctx)
 	}
 
 #if 0
-#ifdef VERBOSE
+#ifdef PED_VERBOSE
 	for (old_cluster=2; old_cluster < old_fs_info->cluster_count+2;
 	     old_cluster++) {
 		if (fat_table_is_available (old_fs_info->fat, old_cluster))
@@ -410,7 +449,7 @@ fat_construct_new_fat (FatOpContext* ctx)
 			fat_table_get (new_fs_info->fat,
 				       ctx->remap [old_cluster]));
 	}
-#endif /* VERBOSE */
+#endif /* PED_VERBOSE */
 #endif
 
 	if (old_fs_info->fat_type == FAT_TYPE_FAT32
@@ -706,6 +745,7 @@ static int
 resize_context_assimilate (FatOpContext* ctx)
 {
 	FatSpecific*	old_fs_info = FAT_SPECIFIC (ctx->old_fs);
+	FatSpecific*	new_fs_info = FAT_SPECIFIC (ctx->new_fs);
 
 	fat_free_buffers (ctx->old_fs);
 	fat_table_destroy (old_fs_info->fat);
@@ -714,6 +754,9 @@ resize_context_assimilate (FatOpContext* ctx)
 
 	ctx->old_fs->type_specific = ctx->new_fs->type_specific;
 	ctx->old_fs->geom = ctx->new_fs->geom;
+	ctx->old_fs->type = (new_fs_info->fat_type == FAT_TYPE_FAT16)
+				? &fat16_type
+			       	: &fat32_type;
 
 	ped_free (ctx->new_fs);
 
@@ -770,7 +813,7 @@ _copy_hidden_sectors (FatOpContext* ctx)
 }
 
 int
-fat_resize (PedFileSystem* fs, const PedGeometry* geom)
+fat_resize (PedFileSystem* fs, PedGeometry* geom, PedTimer* timer)
 {
 	FatSpecific*	fs_info = FAT_SPECIFIC (fs);
 	FatSpecific*	new_fs_info;
@@ -783,7 +826,7 @@ fat_resize (PedFileSystem* fs, const PedGeometry* geom)
 	new_fs = ctx->new_fs;
 	new_fs_info = FAT_SPECIFIC (new_fs);
 
-	if (!fat_duplicate_clusters (ctx))
+	if (!fat_duplicate_clusters (ctx, timer))
 		goto error_abort_ctx;
 	if (fs_info->fat_type == FAT_TYPE_FAT16
 			&& new_fs_info->fat_type == FAT_TYPE_FAT32) {
@@ -792,6 +835,11 @@ fat_resize (PedFileSystem* fs, const PedGeometry* geom)
 	}
 	if (!fat_construct_new_fat (ctx))
 		goto error_abort_ctx;
+	if (fs_info->fat_type == FAT_TYPE_FAT32
+			&& new_fs_info->fat_type == FAT_TYPE_FAT16) {
+		if (!free_root_dir (ctx))
+			goto error_abort_ctx;
+	}
 	if (!fat_construct_dir_tree (ctx))
 		goto error_abort_ctx;
 	if (!fat_table_write_all (new_fs_info->fat, new_fs))
@@ -816,3 +864,4 @@ error:
 	return 0;
 }
 
+#endif /* !DISCOVER_ONLY */
