@@ -20,36 +20,37 @@
 #include "config.h"
 
 #include <parted/parted.h>
+#include <parted/debug.h>
 
-#include <libintl.h>
 #if ENABLE_NLS
+#  include <libintl.h>
 #  define _(String) dgettext (PACKAGE, String)
 #else
 #  define _(String) (String)
 #endif /* ENABLE_NLS */
 
 int
-ped_geometry_init (PedGeometry* geom, const PedDisk* disk,
+ped_geometry_init (PedGeometry* geom, const PedDevice* dev,
 		   PedSector start, PedSector length)
 {
 	PED_ASSERT (geom != NULL, return 0);
-	PED_ASSERT (disk != NULL, return 0);
+	PED_ASSERT (dev != NULL, return 0);
 
-	geom->disk = (PedDisk*) disk;
+	geom->dev = (PedDevice*) dev;
 	return ped_geometry_set (geom, start, length);
 }
 
 PedGeometry*
-ped_geometry_new (const PedDisk* disk, PedSector start, PedSector length)
+ped_geometry_new (const PedDevice* dev, PedSector start, PedSector length)
 {
 	PedGeometry*	geom;
 
-	PED_ASSERT (disk != NULL, return NULL);
+	PED_ASSERT (dev != NULL, return NULL);
 
 	geom = (PedGeometry*) ped_malloc (sizeof (PedGeometry));
 	if (!geom)
 		goto error;
-	if (!ped_geometry_init (geom, disk, start, length))
+	if (!ped_geometry_init (geom, dev, start, length))
 		goto error_free_geom;
 	return geom;
 
@@ -63,7 +64,7 @@ PedGeometry*
 ped_geometry_duplicate (const PedGeometry* geom)
 {
 	PED_ASSERT (geom != NULL, return NULL);
-	return ped_geometry_new (geom->disk, geom->start, geom->length);
+	return ped_geometry_new (geom->dev, geom->start, geom->length);
 }
 
 PedGeometry*
@@ -72,10 +73,7 @@ ped_geometry_intersect (const PedGeometry* a, const PedGeometry* b)
 	PedSector	start;
 	PedSector	end;
 
-	PED_ASSERT (a != NULL, return NULL);
-	PED_ASSERT (b != NULL, return NULL);
-
-	if (a->disk->dev != b->disk->dev)
+	if (!a || !b || a->dev != b->dev)
 		return NULL;
 
 	start = PED_MAX (a->start, b->start);
@@ -83,7 +81,7 @@ ped_geometry_intersect (const PedGeometry* a, const PedGeometry* b)
 	if (start > end)
 		return NULL;
 
-	return ped_geometry_new (a->disk, start, end - start + 1);
+	return ped_geometry_new (a->dev, start, end - start + 1);
 }
 
 void
@@ -98,8 +96,7 @@ int
 ped_geometry_set (PedGeometry* geom, PedSector start, PedSector length)
 {
 	PED_ASSERT (geom != NULL, return 0);
-	PED_ASSERT (geom->disk != NULL, return 0);
-	PED_ASSERT (geom->disk->dev != NULL, return 0);
+	PED_ASSERT (geom->dev != NULL, return 0);
 
 	if (length < 1) {
 		ped_exception_throw (
@@ -108,7 +105,7 @@ ped_geometry_set (PedGeometry* geom, PedSector start, PedSector length)
 			_("Can't have the end before the start!"));
 		return 0;
 	}
-	if (start < 0 || start + length - 1 >= geom->disk->dev->length) {
+	if (start < 0 || start + length - 1 >= geom->dev->length) {
 		ped_exception_throw (
 			PED_EXCEPTION_ERROR,
 			PED_EXCEPTION_CANCEL,
@@ -141,7 +138,7 @@ ped_geometry_test_overlap (const PedGeometry* a, const PedGeometry* b)
 	PED_ASSERT (a != NULL, return 0);
 	PED_ASSERT (b != NULL, return 0);
 
-	if (a->disk->dev != b->disk->dev)
+	if (a->dev != b->dev)
 		return 0;
 
 	if (a->start < b->start)
@@ -156,7 +153,7 @@ ped_geometry_test_inside (const PedGeometry* a, const PedGeometry* b)
 	PED_ASSERT (a != NULL, return 0);
 	PED_ASSERT (b != NULL, return 0);
 
-	if (a->disk->dev != b->disk->dev)
+	if (a->dev != b->dev)
 		return 0;
 
 	return b->start >= a->start && b->end <= a->end;
@@ -168,7 +165,7 @@ ped_geometry_test_equal (const PedGeometry* a, const PedGeometry* b)
 	PED_ASSERT (a != NULL, return 0);
 	PED_ASSERT (b != NULL, return 0);
 
-	return a->disk->dev == b->disk->dev
+	return a->dev == b->dev
 	       && a->start == b->start
 	       && a->end == b->end;
 }
@@ -199,11 +196,11 @@ ped_geometry_read (const PedGeometry* geom, void* buffer, PedSector start,
 			_("Attempt to read sectors %ld-%ld outside of "
 			  "partition on %s"),
 			(long) start, (long) (start + count - 1),
-			geom->disk->dev->path);
+			geom->dev->path);
 		return exception_status == PED_EXCEPTION_IGNORE;
 	}
 
-	if (!ped_device_read (geom->disk->dev, buffer, real_start, count))
+	if (!ped_device_read (geom->dev, buffer, real_start, count))
 		return 0;
 	return 1;
 }
@@ -212,7 +209,7 @@ int
 ped_geometry_sync (PedGeometry* geom)
 {
 	PED_ASSERT (geom != NULL, return 0);
-	return ped_device_sync (geom->disk->dev);
+	return ped_device_sync (geom->dev);
 }
 
 int
@@ -234,11 +231,11 @@ ped_geometry_write (PedGeometry* geom, const void* buffer, PedSector start,
 			_("Attempt to write sectors %ld-%ld outside of "
 			  "partition on %s"),
 			(long) start, (long) (start + count - 1),
-			geom->disk->dev->path);
+			geom->dev->path);
 		return exception_status == PED_EXCEPTION_IGNORE;
 	}
 
-	if (!ped_device_write (geom->disk->dev, buffer, real_start, count))
+	if (!ped_device_write (geom->dev, buffer, real_start, count))
 		return 0;
 	return 1;
 }
@@ -247,7 +244,8 @@ ped_geometry_write (PedGeometry* geom, const void* buffer, PedSector start,
  */
 PedSector
 ped_geometry_check (PedGeometry* geom, void* buffer, PedSector buffer_size,
-		    PedSector offset, PedSector granularity, PedSector count)
+		    PedSector offset, PedSector granularity, PedSector count,
+		    PedTimer* timer)
 {
 	PedSector	group;
 	PedSector	i;
@@ -256,14 +254,19 @@ ped_geometry_check (PedGeometry* geom, void* buffer, PedSector buffer_size,
 	PED_ASSERT (geom != NULL, return 0);
 	PED_ASSERT (buffer != NULL, return 0);
 
+	ped_timer_reset (timer);
+	ped_timer_set_state_name (timer, _("checking for bad blocks"));
+
 retry:
 	ped_exception_fetch_all();
 	for (group = offset; group < offset + count; group += buffer_size) {
+		ped_timer_update (timer, 1.0 * (group - offset) / count);
 		read_len = PED_MIN (buffer_size, offset + count - group);
 		if (!ped_geometry_read (geom, buffer, group, read_len))
 			goto found_error;
 	}
 	ped_exception_leave_all();
+	ped_timer_update (timer, 1.0);
 	return 0;
 
 found_error:
@@ -288,7 +291,7 @@ ped_geometry_map (const PedGeometry* dst, const PedGeometry* src,
 	PED_ASSERT (dst != NULL, return 0);
 	PED_ASSERT (src != NULL, return 0);
 
-	if (dst->disk != src->disk)
+	if (dst->dev != src->dev)
 		return -1;
 
 	result = src->start + sector - dst->start;
