@@ -51,8 +51,6 @@
 #define MIN_FREESPACE		(1000 * 2)	/* 1000k */
 
 typedef struct {
-	int	started;
-	int	finished;
 	time_t	last_update;
 	time_t	predicted_time_left;
 } TimerContext;
@@ -101,59 +99,33 @@ static TimerContext timer_context;
 
 static void _done (PedDevice* dev);
 
-static void
-wipe_line ()
-{
-	/* yuck */
-	printf (
-"\r                                                                          \r"
-	);
-}
-
 static void
 _timer_handler (PedTimer* timer, void* context)
 {
 	TimerContext*	tcontext = (TimerContext*) context;
-	time_t		new_time_left;
+	int		draw_this_time;
 
 	if (opt_script_mode || !isatty(fileno(stdout)))
 		return;
 
-	if (timer->now == timer->start) {
-		tcontext->finished = 0;
-		tcontext->started = 0;
-	}
-	if (timer->now > timer->start) {
-		tcontext->started = 1;
-		tcontext->last_update = 0;
-	}
-
-	/* only allow the estimated time left to go up once every
-	 * 3 seconds, and only after the first 5 seconds... be nice
-	 * to the epileptic ;)
-	 */
-	new_time_left = timer->predicted_end - timer->now;
-	if (timer->now - tcontext->last_update > 3
-	    || (new_time_left < tcontext->predicted_time_left
-		    && timer->now - timer->start > 5
-		    && timer->now != tcontext->last_update)) {
-		tcontext->predicted_time_left = new_time_left;
+	if (tcontext->last_update != timer->now && timer->now > timer->start) {
+		tcontext->predicted_time_left
+			= timer->predicted_end - timer->now;
 		tcontext->last_update = timer->now;
+		draw_this_time = 1;
+	} else {
+		draw_this_time = 0;
 	}
 
-	if (tcontext->started && !tcontext->finished) {
+	if (draw_this_time) {
 		wipe_line ();
 
-		if (timer->frac >= 1.0) {
-			tcontext->finished = 0;
-		} else {
-			if (timer->state_name)
-				printf ("%s... ", timer->state_name);
-			printf (_("%0.f%%\t(time left %.2d:%.2d)"),
-				100.0 * timer->frac,
-				tcontext->predicted_time_left / 60,
-				tcontext->predicted_time_left % 60);
-		}
+		if (timer->state_name)
+			printf ("%s... ", timer->state_name);
+		printf (_("%0.f%%\t(time left %.2d:%.2d)"),
+			100.0 * timer->frac,
+			tcontext->predicted_time_left / 60,
+			tcontext->predicted_time_left % 60);
 
 		fflush (stdout);
 	}
@@ -162,15 +134,24 @@ _timer_handler (PedTimer* timer, void* context)
 static int
 _partition_warn_busy (PedPartition* part)
 {
+	char* path = ped_partition_get_path (part);
+
 	if (ped_partition_is_busy (part)) {
 		if (ped_exception_throw (
-			PED_EXCEPTION_WARNING,
+			PED_EXCEPTION_ERROR,
 			PED_EXCEPTION_IGNORE_CANCEL,
-			_("Partition is being used."))
+			_("Partition %s is being used."),
+			path)
 				!= PED_EXCEPTION_IGNORE)
-			return 0;
+			goto error_free_path;
 	}
+	ped_free (path);
 	return 1;
+
+error_free_path:
+	ped_free (path);
+error:
+	return 0;
 }
 
 static int
@@ -309,6 +290,8 @@ do_check (PedDevice** dev)
 	part = command_line_get_partition (_("Partition number?"), disk);
 	if (!part)
 		goto error_destroy_disk;
+	if (!_partition_warn_busy (part))
+		goto error_destroy_disk;
 
 	if (!ped_disk_check (disk))
 		goto error_destroy_disk;
@@ -569,7 +552,6 @@ do_mkpart (PedDevice** dev)
 		goto error_destroy_part;
 	if (!ped_disk_add_partition (disk, part, constraint))
 		goto error_destroy_part;
-	ped_constraint_destroy (constraint);
 
 	if (!_solution_check_distant (start, end,
 				      part->geom.start, part->geom.end,
@@ -581,6 +563,7 @@ do_mkpart (PedDevice** dev)
 		goto error_destroy_disk;
 	if (!ped_disk_commit (disk))
 		goto error_destroy_disk;
+	ped_constraint_destroy (constraint);
 	ped_disk_destroy (disk);
 	return 1;
 
@@ -711,7 +694,8 @@ do_move (PedDevice** dev)
 	/* get new target */
 	if (!command_line_get_sector (_("Start?"), *dev, 0, &start))
 		goto error_destroy_constraint;
-	if (!command_line_get_sector (_("End?"), *dev, 0, &end))
+	if (!command_line_get_sector (_("End?"), *dev,
+				      start + old_geom.length - 1, &end))
 		goto error_destroy_constraint;
 
 	/* set / test on "disk" */
@@ -900,10 +884,7 @@ do_print (PedDevice** dev)
 		if (!ped_partition_is_active (part))
 	       		continue;
 
-		if (part->type != PED_PARTITION_FREESPACE)
-			printf ("%-5d ", part->num);
-		else
-			printf ("%5s ", "");
+		printf ("%-5d ", part->num);
 
 		printf ("%10.3f %10.3f  ",
 			(int) part->geom.start * 1.0 / MEGABYTE_SECTORS,
@@ -918,11 +899,6 @@ do_print (PedDevice** dev)
 		if (has_name)
 			printf ("%-22s", ped_partition_get_name (part));
 
-		if (part->type == PED_PARTITION_FREESPACE) {
-			printf ("\n");
-			continue;
-		}
-
 		partition_print_flags (part);
 		printf ("\n");
 	}
@@ -943,58 +919,44 @@ do_quit (PedDevice** dev)
 	exit (0);
 }
 
-static PedPartition*
-_rescue_add_partition (PedDisk* disk, const PedFileSystemType* fs_type,
-		       PedGeometry* geom, PedPartitionType part_type)
+static PedPartitionType
+_disk_get_part_type_for_sector (PedDisk* disk, PedSector sector)
 {
-	PedConstraint*		constraint;
-	PedPartition*		new_part;
+	PedPartition*	extended;
 
-	constraint = ped_constraint_new_from_min (geom);
-	if (!constraint)
-		goto error;
+	extended = ped_disk_extended_partition (disk);
+	if (!extended
+	    || !ped_geometry_test_sector_inside (&extended->geom, sector))
+		return 0;
 
-	new_part = ped_partition_new (disk, part_type, fs_type, geom->start,
-				      geom->end);
-	if (!new_part)
-		goto error_destroy_constraint;
-	if (!ped_disk_add_partition (disk, new_part, constraint))
-		goto error_destroy_new_part;
-	if (!ped_disk_commit (disk))
-		goto error_remove_part;
-	ped_constraint_destroy (constraint);
-	return new_part;
-
-error_remove_part:
-	ped_disk_remove_partition (disk, new_part);
-error_destroy_new_part:
-	ped_partition_destroy (new_part);
-error_destroy_constraint:
-	ped_constraint_destroy (constraint);
-error:
-	return 0;
+	return PED_PARTITION_LOGICAL;
 }
 
-PedPartition*
-_rescue_probe_geom (PedPartition* freespace, PedGeometry* probe_region)
+static int
+_rescue_add_partition (PedPartition* part)
 {
 	const PedFileSystemType*	fs_type;
 	PedGeometry*			probed;
 	PedExceptionOption		ex_opt;
-	PedPartitionType		part_type;
+	PedConstraint*			constraint;
 
-	part_type = freespace->type & PED_PARTITION_LOGICAL;
-
-	fs_type = ped_file_system_probe (probe_region);
+	fs_type = ped_file_system_probe (&part->geom);
 	if (!fs_type)
-		return freespace;
-	probed = ped_file_system_probe_specific (fs_type, probe_region);
+		return 0;
+	probed = ped_file_system_probe_specific (fs_type, &part->geom);
 	if (!probed)
-		return freespace;
+		return 0;
 
-	/* assert this? */
-	if (!ped_geometry_test_inside (probe_region, probed))
-		return freespace;
+	if (!ped_geometry_test_inside (&part->geom, probed))
+		return 0;
+
+	constraint = ped_constraint_new_from_min_max (probed, &part->geom);
+	if (!ped_disk_set_partition_geom (part->disk, part, constraint,
+					  probed->start, probed->end)) {
+		ped_constraint_destroy (constraint);
+		return 0;
+	}
+	ped_constraint_destroy (constraint);
 
 	ex_opt = ped_exception_throw (
 		PED_EXCEPTION_INFORMATION,
@@ -1002,135 +964,90 @@ _rescue_probe_geom (PedPartition* freespace, PedGeometry* probe_region)
 		_("A %s %s partition was found at %.3fMb -> %.3fMb.  "
 		  "Do you want to add it to the partition table?"),
 		fs_type->name,
-		ped_partition_type_get_name (part_type),
+		ped_partition_type_get_name (part->type),
 		(int) probed->start * 1.0 / MEGABYTE_SECTORS,
 		(int) probed->end * 1.0 / MEGABYTE_SECTORS);
 
-	if (ex_opt == PED_EXCEPTION_CANCEL)
-		return NULL;
-	if (ex_opt == PED_EXCEPTION_YES
-			|| ex_opt == PED_EXCEPTION_UNHANDLED)
-		return _rescue_add_partition (freespace->disk, fs_type, probed,
-					      part_type);
-	return freespace;
-}
-
-static PedPartition*
-_rescue_align_geom (PedPartition* freespace, PedGeometry* probe_region)
-{
-	PedDisk*		disk = freespace->disk;
-	PedPartitionType	part_type;
-	PedPartition*		new_part;
-	PedConstraint*		constraint;
-
-	part_type = freespace->type & PED_PARTITION_LOGICAL;
-	constraint = ped_constraint_new_from_max (&freespace->geom);
-
-	new_part = ped_partition_new (disk, part_type, NULL,
-				      probe_region->start, probe_region->end);
-	if (!new_part)
-		goto error_destroy_constraint;
-	if (!ped_disk_add_partition (disk, new_part, constraint))
-		goto error_destroy_new_part;
-	ped_geometry_init (probe_region, new_part->geom.dev,
-			   new_part->geom.start, new_part->geom.length);
-	ped_disk_remove_partition (disk, new_part);
-	return ped_disk_get_partition_by_sector (disk, probe_region->start);
-
-error_destroy_new_part:
-	ped_partition_destroy (new_part);
-error_destroy_constraint:
-	ped_constraint_destroy (constraint);
-error:
-	return NULL;
-}
-
-static PedPartition*
-_rescue_probe_freespace (PedPartition* freespace, PedGeometry* probe_region)
-{
-	PedGeometry		sub_region;
-	PedSector		start;
-	PedSector		start_lower_bound;
-	PedSector		start_upper_bound;
-	PedSector		step_size;
-	PedSector		old_start;
-	int			last_pass;
-
-	ped_geometry_init (&sub_region, probe_region->dev,
-			   probe_region->start, probe_region->length);
-
-	start_lower_bound = probe_region->start;
-	start_upper_bound = probe_region->end - 4;	/* min of 2k... */
-	step_size = start_upper_bound - start_lower_bound;
-
-	last_pass = 0;
-	for (step_size = start_upper_bound - start_lower_bound;
-	     step_size && !last_pass; step_size /= 2) {
-		old_start = -1;
-		for (start = start_lower_bound + step_size / 2;
-		     start < start_upper_bound;
-		     start += step_size) {
-			PedPartition*	result;
-
-			ped_geometry_set_start (&sub_region, start);
-
-			freespace = _rescue_align_geom (freespace, &sub_region);
-			if (!freespace)
-				return NULL;
-			if (sub_region.start == old_start)
-				last_pass = 1;
-
-			result = _rescue_probe_geom (freespace, &sub_region);
-			if (!result)
-				return NULL;
-			if (! (result->type & PED_PARTITION_FREESPACE) )
-				return result;
-
-			old_start = sub_region.start;
-		}
+	switch (ex_opt) {
+		case PED_EXCEPTION_CANCEL: return -1;
+		case PED_EXCEPTION_NO: return 0;
 	}
 
-	return freespace;
+	ped_partition_set_system (part, fs_type);
+	ped_disk_commit (part->disk);
+	return 1;
 }
 
+/* hack: we only iterate through the start, since most (all) fs's have their
+ * superblocks at the start.  We'll need to change this if we generalize
+ * for RAID, or something...
+ */
 static int
-_rescue_pass (PedDisk* disk, PedGeometry* probe_region, int thorough)
+_rescue_pass (PedDisk* disk, PedGeometry* start_range, PedGeometry* end_range)
 {
-	PedPartition*	walk;
-	PedPartition*	next;
-	PedGeometry*	sub_region;
+	PedSector		start;
+	PedGeometry		start_geom_exact;
+	PedGeometry		entire_dev;
+	PedConstraint		constraint;
+	PedPartition*		part;
+	PedPartitionType	part_type;
 
-	for (walk = ped_disk_next_partition (disk, NULL); walk; walk = next) {
-		next = ped_disk_next_partition (disk, walk);
+	part_type = _disk_get_part_type_for_sector (
+			disk, (start_range->start + end_range->end) / 2);
 
-		if (! (walk->type & PED_PARTITION_FREESPACE))
+	ped_geometry_init (&entire_dev, disk->dev, 0, disk->dev->length);
+
+	ped_timer_reset (timer);
+	ped_timer_set_state_name (timer, _("searching for file systems"));
+	for (start = start_range->start; start <= start_range->end; start++) {
+		ped_timer_update (timer, 1.0 * (start - start_range->start)
+					 / start_range->length);
+
+		ped_geometry_init (&start_geom_exact, disk->dev, start, 1);
+		ped_constraint_init (
+			&constraint, ped_alignment_any, ped_alignment_any,
+			&start_geom_exact, &entire_dev,
+			1, disk->dev->length);
+		part = ped_partition_new (disk, part_type, NULL, start,
+				end_range->start + end_range->length / 2);
+		if (!part) {
+			ped_constraint_done (&constraint);
 			continue;
-
-		sub_region = ped_geometry_intersect (&walk->geom, probe_region);
-		if (!sub_region)
-			continue;
-		walk = thorough
-			? _rescue_probe_freespace (walk, sub_region)
-			: _rescue_probe_geom (walk, sub_region);
-		ped_geometry_destroy (sub_region);
-
-		if (!walk)
-			return 0;
-
-		/* big nasty hack: if we just rescued a partition, and there
-		 * is free space before it, then we must probe it...
-		 * 	FIXME: what's a better design for this?
-		 */
-		if (! (walk->type & PED_PARTITION_FREESPACE)
-			&& walk->prev
-			&& walk->prev->type & PED_PARTITION_FREESPACE) {
-			next = walk->prev;
-		} else {
-			next = ped_disk_next_partition (disk, walk);
 		}
+
+		ped_exception_fetch_all ();
+		if (ped_disk_add_partition (disk, part, &constraint)) {
+			ped_exception_leave_all ();
+			switch (_rescue_add_partition (part)) {
+			case 1:
+				ped_constraint_done (&constraint);
+				return 1;
+
+			case 0:
+				ped_disk_remove_partition (disk, part);
+				break;
+
+			case -1:
+				goto error_remove_partition;;
+			}
+		} else {
+			ped_exception_leave_all ();
+		}
+		ped_partition_destroy (part);
+		ped_constraint_done (&constraint);
 	}
+	ped_timer_update (timer, 1.0);
 
 	return 1;
+
+error_remove_partition:
+	ped_disk_remove_partition (disk, part);
+error_partition_destroy:
+	ped_partition_destroy (part);
+error_constraint_done:
+	ped_constraint_done (&constraint);
+error:
+	return 0;
 }
 
 static int
@@ -1139,24 +1056,30 @@ do_rescue (PedDevice** dev)
 	PedDisk*		disk;
 	PedSector		start;
 	PedSector		end;
-	PedGeometry		probe_region;
+	PedSector		fuzz;
+	PedGeometry		probe_start_region;
+	PedGeometry		probe_end_region;
 
 	disk = ped_disk_new (*dev);
 	if (!disk)
 		goto error;
 
-	if (!command_line_get_sector (_("Start?"), *dev, 1, &start))
+	if (!command_line_get_sector (_("Start?"), *dev, 0, &start))
 		goto error_destroy_disk;
-	if (!command_line_get_sector (_("End?"), *dev, (*dev)->length - 1, &end))
+	if (!command_line_get_sector (_("End?"), *dev, 0, &end))
 		goto error_destroy_disk;
 
-	ped_geometry_init (&probe_region, *dev, start, end - start + 1);
+	fuzz = PED_MAX (PED_MIN ((end - start) / 10, 1024),
+		        MEGABYTE_SECTORS * 16);
 
-	if (!_rescue_pass (disk, &probe_region, 0))
-		goto error_destroy_disk;
+	ped_geometry_init (&probe_start_region, *dev,
+			   PED_MAX(start - fuzz, 0),
+			   PED_MIN(2 * fuzz, (*dev)->length - (start - fuzz)));
+	ped_geometry_init (&probe_end_region, *dev,
+			   PED_MAX(end - fuzz, 0),
+			   PED_MIN(2 * fuzz, (*dev)->length - (end - fuzz)));
 
-/* should make this optional!  Mac users will hate this! */
-	if (!_rescue_pass (disk, &probe_region, 1))
+	if (!_rescue_pass (disk, &probe_start_region, &probe_end_region))
 		goto error_destroy_disk;
 
 	ped_disk_destroy (disk);
@@ -1488,7 +1411,11 @@ NULL),
 		str_list_create (
 _("print [MINOR]                 display the partition table, or a partition"),
 NULL),
-		NULL));
+		str_list_create (
+_("Without arguments, print displays the entire partition table.  If a\n"
+  "partition number is given, then more detailed information is displayed\n"
+  "about that partition.\n"),
+NULL)));
 	
 	command_register (commands, command_create (
 		str_list_create_unique ("quit", _("quit"), NULL),
@@ -1502,7 +1429,7 @@ NULL),
 		str_list_create_unique ("rescue", _("rescue"), NULL),
 		do_rescue,
 		str_list_create (
-_("rescue START END              rescue lost partitions between START and END"),
+_("rescue START END              rescue a lost partition near START and END"),
 NULL),
 		NULL));
 
@@ -1657,7 +1584,7 @@ _init (int* argc_ptr, char*** argv_ptr)
 	timer = ped_timer_new (_timer_handler, &timer_context);
 	if (!timer)
 		goto error_done_commands;
-	timer_context.started = timer_context.finished = 0;
+	timer_context.last_update = 0;
 
 	return dev;
 
@@ -1681,7 +1608,7 @@ _done (PedDevice* dev)
 		  "rebooting.  Read section 4 of the Parted User "
 		  "documentation for more information."));
 	}
-	if (dev->type != PED_DEVICE_FILE) {
+	if (dev->type != PED_DEVICE_FILE && !opt_script_mode) {
 		ped_exception_throw (
 			PED_EXCEPTION_INFORMATION, PED_EXCEPTION_OK,
 			_("Don't forget to update /etc/fstab, if "
