@@ -121,6 +121,19 @@ screen_width ()
 	return width;
 }
 
+void
+wipe_line ()
+{
+	if (opt_script_mode)
+		return;
+
+	/* yuck */
+	printf (
+"\r                                                                          \r"
+	);
+}
+
+
 #ifdef HAVE_LIBREADLINE
 /* returns matching commands for text */
 static char*
@@ -185,6 +198,7 @@ _readline (const char* prompt, const StrList* possibilities)
 	if (sigsetjmp (readline_state.jmp_state, 1))
 		return NULL;
 
+	wipe_line ();
 #ifdef HAVE_LIBREADLINE
 	if (!opt_script_mode) {
 		/* XXX: why isn't prompt const? */
@@ -232,6 +246,8 @@ _print_exception_text (PedException* ex)
 {
 	StrList*		text;
 
+	wipe_line ();
+
 	if (ex->type == PED_EXCEPTION_BUG) {
 		printf (bug_msg, VERSION);
 		text = str_list_create ("\n", ex->message, "\n", NULL);
@@ -507,7 +523,7 @@ command_line_get_integer (const char* prompt, int def, int* result)
 	return valid;
 }
 
-int
+PedSector
 command_line_get_sector (const char* prompt, PedDevice* dev,
 			 PedSector def, PedSector* result)
 {
